"""Interaction and event mixin for AcherionDesigner."""

# pyright: reportAttributeAccessIssue=false

from __future__ import annotations

import json
from typing import Any

from nicegui import ui

from acherion.constants import (
    _DROP_X_OFFSET,
    _DROP_Y_OFFSET,
    _MANUAL_X,
    _MANUAL_Y,
)
from acherion.registry import (
    is_acherion_manual_add_kind,
)
from acherion.model import (
    AcherionNode,
    _template_title,
)


class _DesignerInteractionsMixin:
    """Selection, drag, grouping, and JS event wiring helpers."""

    _drag_node_id: str | None = None
    _ctx_menu_node_id: str | None = None
    _selected_connection_id: str | None = None

    def _node_by_id(
        self: Any,
        node_id: str | None,
    ) -> AcherionNode | None:
        if not node_id:
            return None
        pure_id = node_id.split('@')[0] if '@' in node_id else node_id
        return next(
            (node for node in self._graph.nodes if node.node_id == pure_id),
            None,
        )

    def _source_label(self: Any, source_id: str | None) -> str:
        if not source_id:
            return 'Unconnected'
        node = self._node_by_id(source_id)
        if node is None:
            return 'Unconnected'
        base = node.title or _template_title(node.kind)
        if source_id and '@' in source_id:
            pin_idx = int(source_id.split('@', 1)[1])
            specs = self._output_pin_specs(node)
            if pin_idx < len(specs):
                return f'{base} ({specs[pin_idx]["label"]})'
        return base

    @staticmethod
    def _connection_id(
        target_node_id: str,
        pin_id: str,
        *,
        source_id: str = '',
    ) -> str:
        if pin_id == 'exec_source' and source_id:
            return f'{target_node_id}@@{pin_id}@@{source_id}'
        return f'{target_node_id}@@{pin_id}'

    @staticmethod
    def _split_connection_id(
        connection_id: str,
    ) -> tuple[str, str, str] | None:
        if '@@' not in connection_id:
            return None
        parts = connection_id.split('@@')
        if len(parts) == 2:
            target_id, pin_id = parts
            return (target_id, pin_id, '')
        if len(parts) == 3:
            target_id, pin_id, source_id = parts
            return (target_id, pin_id, source_id)
        return None

    def _start_drag_node(self: Any, node_id: str, event: Any = None) -> None:
        self._drag_node_id = node_id
        args = dict((event.args if event is not None else None) or {})
        self._drag_offset_x = int(args.get('offset_x') or _DROP_X_OFFSET)
        self._drag_offset_y = int(args.get('offset_y') or _DROP_Y_OFFSET)
        if self._canvas_el is not None:
            self._canvas_el.classes(add='ach-canvas-dragging')
        self._update_hint()

    def _finish_drag_node(self: Any, event: Any) -> None:
        """Persist final drag positions from JS mouseup handler."""
        args = dict(event.args or {})

        if 'rubber_node_ids' in args:
            valid = {node.node_id for node in self._graph.nodes}
            new_sel = {
                node_id
                for node_id in (args['rubber_node_ids'] or [])
                if node_id in valid
            }
            if args.get('add'):
                self._selected_node_ids.update(new_sel)
            else:
                self._selected_node_ids = new_sel
            self._notify_change()
            self._focus_canvas_shortcuts()
            return

        node_id = str(args.get('node_id') or self._drag_node_id or '')
        self._drag_node_id = None
        if self._canvas_el is not None:
            self._canvas_el.classes(remove='ach-canvas-dragging')
        if not node_id:
            return
        dragged = self._node_by_id(node_id)
        if dragged is None:
            return
        raw_left = args.get('left')
        raw_top = args.get('top')
        if raw_left is None or raw_top is None:
            self._update_hint()
            return
        left, top = self._snap_grid_point(raw_left, raw_top)
        dragged.params['x'] = left
        dragged.params['y'] = top
        dragged.params['dock'] = 'free'
        dragged.params['manual_position'] = True
        moved_nodes: list[tuple[AcherionNode, int, int]] = [
            (dragged, left, top)
        ]
        for move in list(args.get('group_moves') or []):
            group_id = str(move.get('node_id') or '')
            group_left = move.get('left')
            group_top = move.get('top')
            if not group_id or group_left is None or group_top is None:
                continue
            group_node = self._node_by_id(group_id)
            if group_node is None:
                continue
            group_node.params['x'] = int(group_left)
            group_node.params['y'] = int(group_top)
            group_node.params['dock'] = 'free'
            group_node.params['manual_position'] = True
            moved_nodes.append(
                (group_node, int(group_left), int(group_top))
            )
        for moved_node, moved_left, moved_top in moved_nodes:
            self._assign_node_to_containing_function_box(
                moved_node,
                moved_left,
                moved_top,
            )
        self._notify_change()

    def _toggle_node_selection(self: Any, event: Any) -> None:
        """Toggle one node or select whole group on normal click."""
        args = dict(event.args or {})
        node_id = str(args.get('node_id') or '')
        toggle = bool(args.get('toggle'))
        valid = {node.node_id for node in self._graph.nodes}
        if node_id not in valid:
            return

        if toggle:
            if node_id in self._selected_node_ids:
                self._selected_node_ids.discard(node_id)
            else:
                self._selected_node_ids.add(node_id)
            self._notify_change()
            self._focus_canvas_shortcuts()
            return

        node = self._node_by_id(node_id)
        if node is None:
            return
        group_ids = self._selection_ids_for_node(node)
        if (
            group_ids == self._selected_node_ids
            and self._selected_connection_id is None
        ):
            self._focus_canvas_shortcuts()
            return
        self._selected_connection_id = None
        self._selected_node_ids = group_ids
        self._notify_change()
        self._focus_canvas_shortcuts()

    def _selection_ids_for_node(self: Any, node: AcherionNode) -> set[str]:
        group_name = str(node.params.get('group') or '').strip()
        if not group_name:
            return {node.node_id}
        return {
            current.node_id
            for current in self._graph.nodes
            if str(current.params.get('group') or '').strip() == group_name
        }

    def _handle_node_contextmenu(self: Any, event: Any) -> None:
        """Show context menu for right-clicked node."""
        args = dict(event.args or {})
        node_id = str(args.get('node_id') or '')
        valid = {node.node_id for node in self._graph.nodes}
        if node_id not in valid:
            return
        if node_id not in self._selected_node_ids:
            node = self._node_by_id(node_id)
            if node is None:
                return
            self._selected_node_ids = self._selection_ids_for_node(node)
        self._ctx_menu_node_id = node_id
        self._ctx_menu_cx = int(args.get('cx') or 0)
        self._ctx_menu_cy = int(args.get('cy') or 0)
        self._reset_context_menu_queries()
        self.refresh()
        self._update_hint()

    def _reset_context_menu_queries(self: Any) -> None:
        self._ctx_menu_query = ''
        self._ctx_align_query = ''

    def _ctx_dismiss(self: Any) -> None:
        """Dismiss context menu without changing selection."""
        self._ctx_menu_node_id = None
        self._reset_context_menu_queries()
        if self._ctx_container_el is not None:
            self._ctx_container_el.clear()
        frame_dom_id = getattr(self, '_frame_dom_id', '')
        if frame_dom_id:
            self._run_client_javascript(
                f'(function(){{'
                f'const frame=document.getElementById({frame_dom_id!r});'
                f'if(!frame)return;'
                f'frame.querySelectorAll(".ach-ctx-menu,.ach-ctx-backdrop")'
                f'.forEach((el) => el.remove());'
                f'}})()'
            )
        self._update_hint()

    def _next_default_group_name(self: Any) -> str:
        """Return next available default group label."""
        index = 1
        while True:
            name = f'Group {index}'
            if name not in self._graph.groups:
                return name
            index += 1

    def _next_default_function_name(self: Any) -> str:
        """Return next available default function-box title."""
        used = {
            str(node.params.get('function_name') or '')
            for node in self._manual_nodes()
            if self._is_function_box(node)
        }
        index = 1
        while True:
            title = f'Function {index}'
            if self._sanitize_identifier(title, 'function_box') not in used:
                return title
            index += 1

    def _selected_group_name(self: Any) -> str:
        """Return shared selected group name when selection has exactly one."""
        groups = {
            str(node.params.get('group') or '').strip()
            for node in self._graph.nodes
            if node.node_id in self._selected_node_ids
            and str(node.params.get('group') or '').strip()
        }
        return next(iter(groups)) if len(groups) == 1 else ''

    def _rename_group(
        self: Any,
        old_name: str,
        new_name: str,
    ) -> tuple[bool, str]:
        """Rename a group while preserving colour and membership."""
        old_name = old_name.strip()
        new_name = new_name.strip()
        if not old_name or old_name not in self._graph.groups:
            return (False, 'Group no longer exists.')
        if not new_name:
            return (False, 'Enter a group name.')
        if new_name != old_name and new_name in self._graph.groups:
            return (False, f'Group {new_name} already exists.')
        if new_name == old_name:
            return (True, f'Group name remains {new_name}.')
        colour = self._graph.groups.pop(old_name)
        self._graph.groups[new_name] = colour
        for node in self._graph.nodes:
            if str(node.params.get('group') or '').strip() == old_name:
                node.params['group'] = new_name
        self._notify_change()
        return (True, f'Renamed group to {new_name}.')

    def _open_new_group_dialog(self: Any) -> None:
        """Create a group from selection using next default name."""
        node_ids = set(self._selected_node_ids)
        self._ctx_dismiss()
        if not node_ids:
            self._notify_ui('Select nodes first.', type='warning')
            return
        name = self._next_default_group_name()
        self._create_group(name, node_ids)
        self._focus_canvas_shortcuts()
        self._notify_ui(f'Created group {name}.', type='positive')

    def _open_rename_group_dialog(self: Any) -> None:
        """Open rename dialog for currently selected group."""
        group_name = self._selected_group_name()
        self._ctx_dismiss()
        if not group_name:
            self._notify_ui(
                'Select nodes from one group first.',
                type='warning',
            )
            return
        if self._overlay_host_el is None:
            return
        with self._overlay_host_el:
            dlg = ui.dialog()
            with dlg, ui.card().style(
                'background:#000000; border:1px solid #2f3336;'
                ' border-radius:12px; padding:20px; min-width:320px;'
            ):
                ui.label('Rename group').classes('text-base font-bold oe-text')
                name_input = ui.input('Group name', value=group_name).classes(
                    'w-full'
                )
                with ui.row().classes('justify-end gap-2 w-full pt-2'):
                    ui.button('Close', on_click=dlg.close).props('flat')

                    def _save() -> None:
                        ok, message = self._rename_group(
                            group_name,
                            str(name_input.value or ''),
                        )
                        self._notify_ui(
                            message,
                            type='positive' if ok else 'warning',
                        )
                        if ok:
                            dlg.close()

                    ui.button('Save', on_click=_save).classes(
                        'oe-btn-primary'
                    )
        dlg.open()

    def _open_extract_function_dialog(self: Any) -> None:
        """Extract selection into function box with default name."""
        anchor_node_id = self._ctx_menu_node_id
        self._ctx_dismiss()
        ok, message = self._extract_function_from_selection(
            self._next_default_function_name(),
            return_node_id=anchor_node_id,
        )
        self._notify_ui(
            message,
            type='positive' if ok else 'warning',
        )

    def _ctx_add_to_group(self: Any, name: str) -> None:
        """Add selected nodes to existing group."""
        node_ids = set(self._selected_node_ids)
        self._ctx_dismiss()
        self._add_nodes_to_group(name, node_ids)
        self._focus_canvas_shortcuts()

    def _ctx_remove_from_group(self: Any) -> None:
        """Remove selected nodes from their group."""
        node_ids = set(self._selected_node_ids)
        self._ctx_dismiss()
        self._remove_nodes_from_group(node_ids)
        self._focus_canvas_shortcuts()

    def _ctx_apply_layout_command(self: Any, command: str) -> None:
        """Apply one alignment/distribution command from the context menu."""
        self._ctx_dismiss()
        ok, message = self._layout_selected_nodes(command)
        self._notify_ui(
            message,
            type='positive' if ok else 'warning',
        )

    def _ctx_delete_selection(self: Any) -> None:
        """Delete selected nodes via context menu."""
        to_delete = set(self._selected_node_ids)
        self._ctx_dismiss()
        self._selected_node_ids.clear()
        self._delete_nodes_batch(to_delete)

    def _handle_viewport_mousemove(self: Any, event: Any) -> None:
        args = dict(event.args or {})
        if str(args.get('action') or '') != 'start':
            return
        node_id = str(args.get('node_id') or '')
        if not node_id:
            return
        self._start_drag_node(node_id, event)

    def _handle_canvas_click(self: Any, event: Any) -> None:
        args = dict(event.args or {})
        action = str(args.get('action') or '')
        if action == 'select':
            self._select_connection(str(args.get('connection_id') or ''))
            return
        if action == 'clear':
            ctx_was_open = self._ctx_menu_node_id is not None
            self._ctx_menu_node_id = None
            changed = (
                ctx_was_open
                or self._selected_connection_id is not None
                or bool(self._selected_node_ids)
            )
            self._selected_connection_id = None
            self._selected_node_ids.clear()
            if changed:
                self.refresh()
            self._update_hint()

    def _handle_canvas_context(self: Any, event: Any) -> None:
        connection_id = str((event.args or {}).get('connection_id') or '')
        if connection_id:
            self._delete_connection(connection_id)

    def _handle_canvas_drop(self: Any, event: Any) -> None:
        args = dict(event.args or {})
        spec_data = str(args.get('spec') or '').strip()
        raw_x = args.get('x')
        raw_y = args.get('y')
        x = _MANUAL_X if raw_x in (None, '') else int(str(raw_x))
        y = _MANUAL_Y if raw_y in (None, '') else int(str(raw_y))
        x, y = self._snap_grid_point(x, y)
        if spec_data:
            try:
                spec = json.loads(spec_data)
            except json.JSONDecodeError:
                spec = None
            if isinstance(spec, dict):
                self._add_variable_node_at_position(spec, x, y)
                self._focus_canvas_shortcuts()
                return
        kind = str(args.get('kind') or '')
        if not is_acherion_manual_add_kind(kind):
            return
        self._add_node_at_position(kind, x, y)
        self._focus_canvas_shortcuts()

    def _handle_canvas_key(self: Any, event: Any) -> None:
        args = dict(event.args or {})
        shortcut_id = str(args.get('shortcut_id') or '')
        key = str(args.get('key') or '')
        raw_anchor_x = args.get('anchor_x')
        raw_anchor_y = args.get('anchor_y')
        anchor_x = None if raw_anchor_x in (None, '') else int(raw_anchor_x)
        anchor_y = None if raw_anchor_y in (None, '') else int(raw_anchor_y)
        if shortcut_id == 'copy_selection':
            ok, message = self._copy_selection_to_clipboard()
            self._notify_ui(
                message,
                type='positive' if ok else 'warning',
            )
            return
        if shortcut_id == 'paste_selection':
            ok, message = self._paste_copied_nodes(
                anchor_x=anchor_x,
                anchor_y=anchor_y,
            )
            self._notify_ui(
                message,
                type='positive' if ok else 'warning',
            )
            return
        if shortcut_id == 'undo_selection':
            ok, message = self._undo_graph_change()
            self._notify_ui(
                message,
                type='positive' if ok else 'warning',
            )
            return
        if shortcut_id == 'redo_selection':
            ok, message = self._redo_graph_change()
            self._notify_ui(
                message,
                type='positive' if ok else 'warning',
            )
            return
        if shortcut_id == 'delete_selection_primary':
            if self._selected_node_ids:
                to_delete = set(self._selected_node_ids)
                self._selected_node_ids.clear()
                self._delete_nodes_batch(to_delete)
            else:
                self._delete_selected_connection()
            return
        if shortcut_id == 'clear_selection' or key == 'Escape':
            if self._selected_node_ids:
                self._selected_node_ids.clear()
                self.refresh()
                self._update_hint()
            elif self._selected_connection_id is not None:
                self._selected_connection_id = None
                self.refresh()
                self._update_hint()

    def _attach_viewport_events(self: Any, viewport: Any) -> None:
        viewport.on(
            'mousedown',
            lambda _e: None,
            js_handler=(
                '(e) => {'
                'const vp = e.currentTarget;'
                'vp.dataset.cursorClientX = String(e.clientX);'
                'vp.dataset.cursorClientY = String(e.clientY);'
                'if (e.button !== 0) return;'
                'if (e.target.closest(".ach-node")) return;'
                'if (e.target.closest(".ach-link-hitbox")) return;'
                'if (e.target.closest(".ach-palette-shell")) return;'
                'const h = window.__oeAcherion;'
                'if (h && h.matchesShortcut(e, "box_select_add", "drag")) {'
                'h.ensureViewportState(vp);'
                'const pt = h.worldPoint(vp, e.clientX, e.clientY);'
                'vp.dataset.rubberActive = "1";'
                'vp.dataset.rubberStartX = String(pt.x);'
                'vp.dataset.rubberStartY = String(pt.y);'
                'vp.dataset.rubberShift = "1";'
                'vp.classList.add("ach-shell-selecting");'
                'return;'
                '}'
                'if (h && h.matchesShortcut(e, "box_select", "drag")) {'
                'h.ensureViewportState(vp);'
                'const pt = h.worldPoint(vp, e.clientX, e.clientY);'
                'vp.dataset.rubberActive = "1";'
                'vp.dataset.rubberStartX = String(pt.x);'
                'vp.dataset.rubberStartY = String(pt.y);'
                'vp.dataset.rubberShift = "0";'
                'vp.classList.add("ach-shell-selecting");'
                'return;'
                '}'
                'vp.dataset.panActive = "1";'
                'vp.dataset.panLastX = String(e.clientX);'
                'vp.dataset.panLastY = String(e.clientY);'
                'vp.classList.add("ach-shell-panning");'
                '}'
            ),
        )
        viewport.on(
            'mousemove',
            self._handle_viewport_mousemove,
            js_handler=(
                '(e) => {'
                'const vp = e.currentTarget;'
                'vp.dataset.cursorClientX = String(e.clientX);'
                'vp.dataset.cursorClientY = String(e.clientY);'
                'const h = window.__oeAcherion;'
                'if (h) h.ensureViewportState(vp);'
                'const cid = vp.dataset.dragCandidateNodeId || "";'
                'if (cid && h) {'
                'const sx = parseFloat(vp.dataset.dragStartClientX || "0");'
                'const sy = parseFloat(vp.dataset.dragStartClientY || "0");'
                'const moved = Math.hypot(e.clientX - sx, e.clientY - sy);'
                'let aid = vp.dataset.dragNodeId || "";'
                'if (!aid && moved >= 4) {'
                'aid = cid;'
                'vp.dataset.dragNodeId = aid;'
                'vp.classList.add("ach-shell-dragging-node");'
                'const stage = h.stage(vp);'
                'const nd = stage ? h.findNode(stage, aid) : null;'
                'if (nd) nd.classList.add("ach-node-dragging");'
                'if (stage && nd) {'
                'const selNds = Array.from(stage.querySelectorAll(".ach-node-selected")).map(n=>n.dataset.nodeId);'
                'if (selNds.includes(aid) && selNds.length > 1) {'
                'const dl=parseFloat(nd.style.left||"0"),dt=parseFloat(nd.style.top||"0");'
                'const grp=selNds.filter(id=>id!==aid).map(id=>{'
                'const gn=h.findNode(stage,id);'
                'return gn?{id,rx:parseFloat(gn.style.left||"0")-dl,ry:parseFloat(gn.style.top||"0")-dt}:null;'
                '}).filter(Boolean);'
                'vp.dataset.dragGroupData=JSON.stringify(grp);'
                '} else if ((nd.dataset.isFunctionBox || "") === "1") {'
                'const fid = nd.dataset.nodeId || aid;'
                'const dl=parseFloat(nd.style.left||"0"),dt=parseFloat(nd.style.top||"0");'
                'const inBox=(n,targetId)=>{'
                'let pid=n.dataset.parentFunctionId||"";'
                'let guard=0;'
                'while(pid && guard<32){'
                'if(pid===targetId)return true;'
                'const parent=h.findNode(stage,pid);'
                'pid=parent?(parent.dataset.parentFunctionId||""):"";'
                'guard+=1;'
                '}'
                'return false;'
                '};'
                'const grp=Array.from(stage.querySelectorAll(".ach-node")).filter(n=>'
                '(n.dataset.nodeId||"")!==aid && inBox(n,fid)).map(n=>({'
                'id:n.dataset.nodeId||"",'
                'rx:parseFloat(n.style.left||"0")-dl,'
                'ry:parseFloat(n.style.top||"0")-dt'
                '})).filter(item=>item.id);'
                'vp.dataset.dragGroupData=JSON.stringify(grp);'
                '} else if ((nd.dataset.groupToken || "")) {'
                'const gt = nd.dataset.groupToken || "";'
                'const dl=parseFloat(nd.style.left||"0"),dt=parseFloat(nd.style.top||"0");'
                'const grp=Array.from(stage.querySelectorAll(".ach-node")).filter(n=>'
                '(n.dataset.groupToken||"")===gt && (n.dataset.nodeId||"")!==aid).map(n=>({'
                'id:n.dataset.nodeId||"",'
                'rx:parseFloat(n.style.left||"0")-dl,'
                'ry:parseFloat(n.style.top||"0")-dt'
                '})).filter(item=>item.id);'
                'vp.dataset.dragGroupData=JSON.stringify(grp);'
                '} else { vp.dataset.dragGroupData="[]"; }'
                '} else { vp.dataset.dragGroupData="[]"; }'
                'emit({action:"start",node_id:aid,'
                'offset_x:Math.round(parseFloat(vp.dataset.dragOffsetX||"0")),'
                'offset_y:Math.round(parseFloat(vp.dataset.dragOffsetY||"0"))});'
                '}'
                'if (aid) {'
                'const pt = h.worldPoint(vp, e.clientX, e.clientY);'
                'const snapped = h.snapPoint('
                'vp,'
                'pt.x - parseFloat(vp.dataset.dragOffsetX||"0"),'
                'pt.y - parseFloat(vp.dataset.dragOffsetY||"0")'
                ');'
                'const l = snapped.x;'
                'const t = snapped.y;'
                'const stage = h.stage(vp);'
                'if (stage) {'
                'h.moveNode(stage, aid, l, t);'
                'const grp=JSON.parse(vp.dataset.dragGroupData||"[]");'
                'grp.forEach(item=>h.moveNode(stage,item.id,l+item.rx,t+item.ry));'
                '}'
                'return;'
                '}'
                '}'
                'if (vp.dataset.rubberActive==="1" && h) {'
                'const pt=h.worldPoint(vp,e.clientX,e.clientY);'
                'h.updateRubberBand(vp,parseFloat(vp.dataset.rubberStartX||"0"),parseFloat(vp.dataset.rubberStartY||"0"),pt.x,pt.y);'
                'return;'
                '}'
                'if (vp.dataset.panActive !== "1") return;'
                'const stage = vp.querySelector(".ach-canvas");'
                'if (!stage) return;'
                'const lx = parseFloat(vp.dataset.panLastX || "0");'
                'const ly = parseFloat(vp.dataset.panLastY || "0");'
                'const dx = e.clientX - lx, dy = e.clientY - ly;'
                'const px = parseFloat(vp.dataset.panX || "0") + dx;'
                'const py = parseFloat(vp.dataset.panY || "0") + dy;'
                'vp.dataset.panX = String(px);'
                'vp.dataset.panY = String(py);'
                'vp.dataset.panLastX = String(e.clientX);'
                'vp.dataset.panLastY = String(e.clientY);'
                'h.applyViewportTransform(vp);'
                '}'
            ),
        )
        finish_js = (
            '(e) => {'
            'const vp = e.currentTarget;'
            'const h = window.__oeAcherion;'
            'const aid = vp.dataset.dragNodeId || "";'
            'const clear = () => {'
            'vp.dataset.dragCandidateNodeId = "";'
            'vp.dataset.dragGroupToken = "";'
            'vp.dataset.dragStartClientX = "";'
            'vp.dataset.dragStartClientY = "";'
            'vp.dataset.dragOffsetX = "";'
            'vp.dataset.dragOffsetY = "";'
            '};'
            'if (vp.dataset.rubberActive === "1" && h) {'
            'const stage = h.stage(vp);'
            'const pt = h.worldPoint(vp, e.clientX, e.clientY);'
            'const x1=parseFloat(vp.dataset.rubberStartX||"0");'
            'const y1=parseFloat(vp.dataset.rubberStartY||"0");'
            'const nodeIds=stage?h.nodesInRect(stage,x1,y1,pt.x,pt.y):[];'
            'h.clearRubberBand(vp);'
            'vp.dataset.rubberActive="";'
            'vp.dataset.suppressNextCanvasClick="1";'
            'vp.classList.remove("ach-shell-selecting");'
            'clear();'
            'emit({rubber_node_ids:nodeIds,add:vp.dataset.rubberShift==="1"});'
            'return;'
            '}'
            'if (aid && h) {'
            'const pt = h.worldPoint(vp, e.clientX, e.clientY);'
            'const snapped = h.snapPoint('
            'vp,'
            'pt.x - parseFloat(vp.dataset.dragOffsetX||"0"),'
            'pt.y - parseFloat(vp.dataset.dragOffsetY||"0")'
            ');'
            'const l = snapped.x;'
            'const t = snapped.y;'
            'const stage = h.stage(vp);'
            'const grpData = JSON.parse(vp.dataset.dragGroupData||"[]");'
            'if (stage) {'
            'h.moveNode(stage, aid, l, t);'
            'const nd = h.findNode(stage, aid);'
            'if (nd) nd.classList.remove("ach-node-dragging");'
            'grpData.forEach(item => {'
            'h.moveNode(stage,item.id,l+item.rx,t+item.ry);'
            'const gnd=h.findNode(stage,item.id);if(gnd)gnd.classList.remove("ach-node-dragging");'
            '});'
            '}'
            'vp.dataset.dragNodeId = "";'
            'vp.classList.remove("ach-shell-dragging-node");'
            'vp.dataset.dragGroupData = "[]";'
            'clear();'
            'const gm = grpData.map(item => ({'
            'node_id:item.id,'
            'left:Math.round(l + item.rx),'
            'top:Math.round(t + item.ry)'
            '}));'
            'emit({node_id:aid,left:Math.round(l),top:Math.round(t),group_moves:gm});'
            'return;'
            '}'
            'clear();'
            'vp.dataset.panActive = "0";'
            'vp.classList.remove("ach-shell-panning");'
            '}'
        )
        viewport.on('mouseup', self._finish_drag_node, js_handler=finish_js)
        viewport.on('mouseleave', self._finish_drag_node, js_handler=finish_js)
        viewport.on(
            'wheel',
            lambda _e: None,
            js_handler=(
                '(e) => {'
                'const vp = e.currentTarget;'
                'const h = window.__oeAcherion;'
                'if (e.target.closest(".ach-palette-shell")) return;'
                'if (vp.dataset.dragNodeId || vp.dataset.dragCandidateNodeId) return;'
                'if (!h || !h.matchesShortcut(e, "zoom_canvas", "wheel")) return;'
                'e.preventDefault();'
                'h.ensureViewportState(vp);'
                'const stage = vp.querySelector(".ach-canvas");'
                'if (!stage) return;'
                'const rect = vp.getBoundingClientRect();'
                'const os = parseFloat(vp.dataset.scale || "1");'
                'const zoom = e.deltaY < 0 ? 1.12 : 0.89;'
                'const ns = Math.max(0.08, Math.min(4.0, os * zoom));'
                'const px = parseFloat(vp.dataset.panX || "0");'
                'const py = parseFloat(vp.dataset.panY || "0");'
                'const wx = (e.clientX - rect.left - px) / os;'
                'const wy = (e.clientY - rect.top - py) / os;'
                'vp.dataset.scale = String(ns);'
                'vp.dataset.panX = String((e.clientX - rect.left) - wx * ns);'
                'vp.dataset.panY = String((e.clientY - rect.top) - wy * ns);'
                'h.applyViewportTransform(vp);'
                '}'
            ),
        )
        viewport.on(
            'dragover',
            lambda _e: None,
            js_handler=(
                '(e) => {'
                "if (!e.dataTransfer.types.includes('text/ach-kind') && !e.dataTransfer.types.includes('text/ach-node-spec')) return;"
                'e.preventDefault();'
                "e.dataTransfer.dropEffect = 'copy';"
                '}'
            ),
        )
        viewport.on(
            'drop',
            self._handle_canvas_drop,
            js_handler=(
                '(e) => {'
                "const spec = e.dataTransfer.getData('text/ach-node-spec');"
                "const kind = e.dataTransfer.getData('text/ach-kind');"
                'if (!spec && !kind) return;'
                'e.preventDefault();'
                'const vp = e.currentTarget;'
                'const h = window.__oeAcherion;'
                'if (!h) return;'
                'const pt = h.worldPoint(vp, e.clientX, e.clientY);'
                'const snapped = h.snapPoint(vp, pt.x, pt.y);'
                'emit({kind: kind, spec: spec, x: Math.round(snapped.x), y: Math.round(snapped.y)});'
                '}'
            ),
        )

    def _attach_canvas_events(self: Any, canvas: Any) -> None:
        canvas.on(
            'click',
            self._handle_canvas_click,
            js_handler=(
                '(e) => {'
                'const vp = e.currentTarget.closest(".ach-shell");'
                'if (vp && vp.dataset.suppressNextCanvasClick === "1") {'
                'vp.dataset.suppressNextCanvasClick = "";'
                'return;'
                '}'
                'if (e.ctrlKey || e.metaKey) return;'
                'const hit = e.target.closest(".ach-link-hitbox");'
                'if (hit) {'
                'e.preventDefault(); e.stopPropagation();'
                'e.currentTarget.focus();'
                'emit({action:"select",'
                'connection_id:hit.dataset.connectionId||""});'
                'return;'
                '}'
                'if (e.target.closest(".ach-node")) return;'
                'emit({action:"clear"});'
                '}'
            ),
        )
        canvas.on(
            'contextmenu.prevent',
            self._handle_canvas_context,
            js_handler=(
                '(e) => {'
                'const hit = e.target.closest(".ach-link-hitbox");'
                'if (!hit) return;'
                'e.preventDefault(); e.stopPropagation();'
                'e.currentTarget.focus();'
                'emit({connection_id:hit.dataset.connectionId||""});'
                '}'
            ),
        )
        canvas.on(
            'keydown',
            self._handle_canvas_key,
            js_handler=(
                '(e) => {'
                'const h = window.__oeAcherion;'
                'if (!h) return;'
                'const tag = (e.target.tagName || "").toUpperCase();'
                'if (tag === "INPUT" || tag === "TEXTAREA" || tag === "SELECT"'
                ' || e.target.isContentEditable) return;'
                'const vp = e.currentTarget.closest(".ach-shell");'
                'let shortcutId = "";'
                'if (h.matchesShortcut(e, "delete_selection_primary", "keyboard")) {'
                'shortcutId = "delete_selection_primary";'
                '} else if (h.matchesShortcut(e, "copy_selection", "keyboard")) {'
                'shortcutId = "copy_selection";'
                '} else if (h.matchesShortcut(e, "paste_selection", "keyboard")) {'
                'shortcutId = "paste_selection";'
                '} else if (h.matchesShortcut(e, "undo_selection", "keyboard")) {'
                'shortcutId = "undo_selection";'
                '} else if (h.matchesShortcut(e, "redo_selection", "keyboard")) {'
                'shortcutId = "redo_selection";'
                '} else if (h.matchesShortcut(e, "clear_selection", "keyboard")) {'
                'shortcutId = "clear_selection";'
                '}'
                'if (!shortcutId) return;'
                'let anchorX = null;'
                'let anchorY = null;'
                'if (shortcutId === "paste_selection" && vp) {'
                'h.ensureViewportState(vp);'
                'const rawClientX = parseFloat(vp.dataset.cursorClientX || "");'
                'const rawClientY = parseFloat(vp.dataset.cursorClientY || "");'
                'if (Number.isFinite(rawClientX) && Number.isFinite(rawClientY)) {'
                'const pt = h.worldPoint(vp, rawClientX, rawClientY);'
                'const snapped = h.snapPoint(vp, pt.x, pt.y);'
                'anchorX = Math.round(snapped.x);'
                'anchorY = Math.round(snapped.y);'
                '}'
                '}'
                'e.preventDefault(); e.stopPropagation();'
                'emit({shortcut_id: shortcutId, key:e.key, anchor_x: anchorX, anchor_y: anchorY});'
                '}'
            ),
        )


"""Core JavaScript helpers for the Acherion render surface."""

from __future__ import annotations

import json

import acherion.preferences as acherion_preferences


_SHORTCUT_CLIENT_CONFIG_JSON = json.dumps(
    acherion_preferences.shortcut_client_config(),
    sort_keys=True,
)


CORE_JS = """
        shortcutConfig() {
            return __ACH_SHORTCUT_CONFIG__;
        },
        defaultPreferences() {
            return this.shortcutConfig().default_preferences || {};
        },
        preferences() {
            if (!this._preferences) {
                this.setPreferences(this.defaultPreferences());
            }
            return this._preferences;
        },
        setPreferences(preferences) {
            const fallback = this.defaultPreferences();
            const raw = preferences || {};
            const appearance = raw.appearance || {};
            const shortcuts = raw.keyboard_shortcuts || {};
            this._preferences = {
                appearance: {
                    code_editor_theme: String(
                        appearance.code_editor_theme ||
                        fallback.appearance.code_editor_theme ||
                        ''
                    ),
                },
                keyboard_shortcuts: {
                    ...fallback.keyboard_shortcuts,
                    ...shortcuts,
                },
            };
        },
        shortcutBinding(identifier) {
            const shortcuts = this.preferences().keyboard_shortcuts || {};
            return String(shortcuts[identifier] || '').trim();
        },
        normalizeShortcutToken(token) {
            const clean = String(token || '').trim().toUpperCase();
            const aliases = this.shortcutConfig().token_aliases || {};
            return String(aliases[clean] || clean);
        },
        normalizeShortcutKey(key) {
            return this.normalizeShortcutToken(key);
        },
        parseShortcutBinding(binding) {
            const shortcutConfig = this.shortcutConfig();
            const modifierTokens = new Set(
                shortcutConfig.modifier_tokens || []
            );
            const mouseInteractions = new Set(
                shortcutConfig.mouse_interactions || []
            );
            const parts = String(binding || '')
                .split('+')
                .map(token => token.trim())
                .filter(Boolean);
            if (!parts.length) return null;
            const modifiers = new Set();
            const tail = this.normalizeShortcutToken(parts[parts.length - 1]);
            let interaction = 'keyboard';
            if (mouseInteractions.has(tail)) {
                interaction = tail.toLowerCase();
            }
            for (const token of parts.slice(0, -1)) {
                const normalized = this.normalizeShortcutToken(token);
                if (!modifierTokens.has(normalized)) {
                    return null;
                }
                modifiers.add(normalized);
            }
            return {
                interaction,
                key: interaction === 'keyboard' ? tail : '',
                modifiers,
            };
        },
        matchesBinding(event, binding, interaction) {
            const parsed = this.parseShortcutBinding(binding);
            if (!parsed || parsed.interaction !== interaction) return false;
            const hasCtrl = !!event.ctrlKey;
            const hasMeta = !!event.metaKey;
            const hasShift = !!event.shiftKey;
            const hasAlt = !!event.altKey;
            const wantsCtrl = parsed.modifiers.has('CTRL');
            const wantsMeta = parsed.modifiers.has('META');
            const wantsShift = parsed.modifiers.has('SHIFT');
            const wantsAlt = parsed.modifiers.has('ALT');
            const wantsMod = parsed.modifiers.has('MOD');
            if (wantsShift !== hasShift) return false;
            if (wantsAlt !== hasAlt) return false;
            if (wantsMod) {
                if (!(hasCtrl || hasMeta)) return false;
                if (wantsCtrl && !hasCtrl) return false;
                if (wantsMeta && !hasMeta) return false;
            } else {
                if (wantsCtrl !== hasCtrl) return false;
                if (wantsMeta !== hasMeta) return false;
            }
            if (!wantsMod && !wantsCtrl && !wantsMeta && (hasCtrl || hasMeta)) {
                return false;
            }
            if (interaction !== 'keyboard') return true;
            return this.normalizeShortcutKey(event.key) === parsed.key;
        },
        matchesShortcut(event, identifier, interaction) {
            return this.matchesBinding(
                event,
                this.shortcutBinding(identifier),
                interaction
            );
        },
        queueSync(viewport) {
            if (!viewport) return;
            if (viewport.__oeVlSyncQueued) return;
            viewport.__oeVlSyncQueued = true;
            requestAnimationFrame(() => {
                viewport.__oeVlSyncQueued = false;
                this.syncViewport(viewport);
            });
        },
        stage(viewport) {
            return viewport ? viewport.querySelector('.ach-canvas') : null;
        },
        origin(stage) {
            if (!stage) return {x: 0, y: 0};
            const styles = getComputedStyle(stage);
            const ox = parseFloat(styles.getPropertyValue('--ach-origin-x') || '0');
            const oy = parseFloat(styles.getPropertyValue('--ach-origin-y') || '0');
            return {
                x: Number.isFinite(ox) ? ox : 0,
                y: Number.isFinite(oy) ? oy : 0,
            };
        },
        chrome(viewport) {
            if (!viewport) return {sidebarWidth: 0, pad: 24};
            const styles = getComputedStyle(viewport);
            const sidebarWidth = parseFloat(
                styles.getPropertyValue('--ach-sidebar-width') || '0'
            );
            const pad = parseFloat(
                styles.getPropertyValue('--ach-stage-pad') || '24'
            );
            return {
                sidebarWidth: Number.isFinite(sidebarWidth) ? sidebarWidth : 0,
                pad: Number.isFinite(pad) ? pad : 24,
            };
        },
        frame(element) {
            return element ? element.closest('.ach-workbench') : null;
        },
        contextSubmenu(item) {
            if (!item) return null;
            return Array.from(item.children || []).find(
                child => child.classList?.contains('ach-ctx-submenu')
            ) || null;
        },
        measureWhileHidden(element, measure) {
            if (!element || typeof measure !== 'function') return null;
            const prevDisplay = element.style.display;
            const prevVisibility = element.style.visibility;
            const prevPointerEvents = element.style.pointerEvents;
            element.style.display = 'block';
            element.style.visibility = 'hidden';
            element.style.pointerEvents = 'none';
            const result = measure();
            element.style.display = prevDisplay;
            element.style.visibility = prevVisibility;
            element.style.pointerEvents = prevPointerEvents;
            return result;
        },
        positionContextSubmenus(menu) {
            if (!menu) return;
            const margin = 12;
            const viewportWidth =
                window.innerWidth || document.documentElement.clientWidth || 0;
            const viewportHeight =
                window.innerHeight || document.documentElement.clientHeight || 0;
            menu.querySelectorAll('.ach-ctx-item-has-submenu').forEach(item => {
                const submenu = this.contextSubmenu(item);
                if (!submenu) return;
                const size = this.measureWhileHidden(submenu, () => {
                    const rect = submenu.getBoundingClientRect();
                    return {width: rect.width, height: rect.height};
                });
                if (!size) return;
                const itemRect = item.getBoundingClientRect();
                const desiredAbsoluteLeft = itemRect.left + itemRect.width - 4;
                const maxLeft = Math.max(
                    margin,
                    viewportWidth - size.width - margin
                );
                const absoluteLeft = Math.max(
                    margin,
                    Math.min(maxLeft, desiredAbsoluteLeft)
                );
                const maxTop = Math.max(
                    margin,
                    viewportHeight - size.height - margin
                );
                const absoluteTop = Math.max(
                    margin,
                    Math.min(maxTop, itemRect.top - 10)
                );
                item.style.setProperty(
                    '--ach-ctx-submenu-left',
                    `${absoluteLeft - itemRect.left}px`
                );
                item.style.setProperty(
                    '--ach-ctx-submenu-top',
                    `${absoluteTop - itemRect.top}px`
                );
            });
        },
        positionContextMenu(menu, anchorX, anchorY) {
            if (!menu) return;
            const margin = 12;
            const viewportWidth =
                window.innerWidth || document.documentElement.clientWidth || 0;
            const viewportHeight =
                window.innerHeight || document.documentElement.clientHeight || 0;
            menu.style.left = `${Math.max(margin, anchorX)}px`;
            menu.style.top = `${Math.max(margin, anchorY)}px`;
            const rect = menu.getBoundingClientRect();
            const maxLeft = Math.max(margin, viewportWidth - rect.width - margin);
            const maxTop = Math.max(
                margin,
                viewportHeight - rect.height - margin
            );
            menu.style.left = `${Math.max(margin, Math.min(maxLeft, anchorX))}px`;
            menu.style.top = `${Math.max(margin, Math.min(maxTop, anchorY))}px`;
            this.positionContextSubmenus(menu);
        },
        palette(frame) {
            return frame ? frame.querySelector('.ach-palette-shell') : null;
        },
""".replace(
    '__ACH_SHORTCUT_CONFIG__',
    _SHORTCUT_CLIENT_CONFIG_JSON,
)

__all__ = ['CORE_JS']


"""Sidebar JavaScript for the Acherion render surface."""

from __future__ import annotations

SIDEBAR_JS = """
        defaultSidebarPaneWidth(frame) {
            if (!frame) return 244;
            const styles = getComputedStyle(frame);
            const width = parseFloat(
                styles.getPropertyValue(
                    '--ach-palette-pane-default-width'
                ) || '244'
            );
            return Number.isFinite(width) ? width : 244;
        },
        sidebarPaneWidth(frame) {
            if (!frame) return 0;
            const styles = getComputedStyle(frame);
            const width = parseFloat(
                styles.getPropertyValue('--ach-palette-pane-width') || '0'
            );
            return Number.isFinite(width)
                ? width : this.defaultSidebarPaneWidth(frame);
        },
        sidebarPaneMinWidth(frame) {
            if (!frame) return 188;
            const styles = getComputedStyle(frame);
            const width = parseFloat(
                styles.getPropertyValue('--ach-palette-pane-min-width') || '188'
            );
            return Number.isFinite(width) ? width : 188;
        },
        sidebarPaneMaxWidth(frame) {
            if (!frame) return 460;
            const min = this.sidebarPaneMinWidth(frame);
            const styles = getComputedStyle(frame);
            const configured = parseFloat(
                styles.getPropertyValue('--ach-palette-pane-max-width') || '460'
            );
            const frameWidth = frame?.clientWidth || window.innerWidth || 0;
            const availableMax = Math.max(
                min,
                frameWidth > 0 ? frameWidth - 260 : configured
            );
            return Math.max(
                min,
                Math.min(
                    Number.isFinite(configured) ? configured : 460,
                    availableMax
                )
            );
        },
        sidebarCollapseThreshold(frame) {
            if (!frame) return 124;
            const styles = getComputedStyle(frame);
            const width = parseFloat(
                styles.getPropertyValue(
                    '--ach-palette-pane-collapse-threshold'
                ) || '124'
            );
            return Number.isFinite(width) ? width : 124;
        },
        syncSidebarState(frame) {
            if (!frame) return;
            const palette = this.palette(frame);
            if (!palette) return;
            const buttons = Array.from(
                palette.querySelectorAll('.ach-activity-button')
            );
            const panes = Array.from(
                palette.querySelectorAll('.ach-sidebar-pane')
            );
            let activeKey = palette.dataset.activeSidebar || '';
            if (!buttons.some(button => button.dataset.sidebarKey === activeKey)) {
                activeKey = buttons[0]?.dataset.sidebarKey || '';
                palette.dataset.activeSidebar = activeKey;
            }
            const paneOpen = this.sidebarPaneWidth(frame) > 0.5;
            palette.dataset.paneOpen = paneOpen ? '1' : '0';
            frame.classList.toggle('ach-sidebar-pane-closed', !paneOpen);
            buttons.forEach(button => {
                const isActive = button.dataset.sidebarKey === activeKey;
                button.classList.toggle('ach-activity-button-active', isActive);
                button.setAttribute(
                    'aria-pressed',
                    isActive ? 'true' : 'false'
                );
            });
            panes.forEach(pane => {
                const isVisible = (
                    paneOpen && pane.dataset.sidebarKey === activeKey
                );
                pane.classList.toggle('ach-sidebar-pane-active', isVisible);
            });
        },
        clampSidebarPaneWidth(frame, widthValue) {
            const min = this.sidebarPaneMinWidth(frame);
            const max = this.sidebarPaneMaxWidth(frame);
            return Math.max(min, Math.min(max, widthValue));
        },
        setSidebarPaneWidth(frame, widthValue, persist = true) {
            if (!frame) return;
            const nextWidth = this.clampSidebarPaneWidth(frame, widthValue);
            frame.style.setProperty(
                '--ach-palette-pane-width',
                `${nextWidth}px`
            );
            if (persist && nextWidth > 0)
                frame.dataset.sidebarExpandedWidth = String(nextWidth);
            this.syncSidebarState(frame);
        },
        closeSidebarPane(frame) {
            if (!frame) return;
            const currentWidth = this.sidebarPaneWidth(frame);
            if (currentWidth > 0) {
                frame.dataset.sidebarExpandedWidth = String(
                    Math.max(this.sidebarPaneMinWidth(frame), currentWidth)
                );
            }
            frame.style.setProperty('--ach-palette-pane-width', '0px');
            this.syncSidebarState(frame);
        },
        openSidebarPane(frame) {
            if (!frame) return;
            const defaultWidth = this.defaultSidebarPaneWidth(frame);
            const storedWidth = parseFloat(
                frame.dataset.sidebarExpandedWidth || String(defaultWidth)
            );
            const nextWidth = Number.isFinite(storedWidth)
                ? storedWidth : defaultWidth;
            this.setSidebarPaneWidth(frame, nextWidth);
        },
        resetSidebarWidth(frame) {
            if (!frame) return;
            const defaultWidth = this.defaultSidebarPaneWidth(frame);
            frame.dataset.sidebarExpandedWidth = String(defaultWidth);
            this.setSidebarPaneWidth(frame, defaultWidth);
        },
        toggleSidebarSection(button) {
            const frame = this.frame(button);
            const palette = this.palette(frame);
            const key = button?.dataset?.sidebarKey || '';
            if (!frame || !palette || !key) return;
            const currentKey = palette.dataset.activeSidebar || '';
            const paneOpen = this.sidebarPaneWidth(frame) > 0.5;
            if (paneOpen && currentKey === key) {
                this.closeSidebarPane(frame);
                return;
            }
            palette.dataset.activeSidebar = key;
            this.openSidebarPane(frame);
        },
        beginSidebarResize(handle, startClientX) {
            const frame = this.frame(handle);
            if (!frame) return;
            this._stopSidebarResize?.();
            const startWidth = this.sidebarPaneWidth(frame);
            document.body?.classList.add('ach-sidebar-resizing');
            const move = (event) => {
                const nextWidth = startWidth + (event.clientX - startClientX);
                if (nextWidth <= this.sidebarCollapseThreshold(frame)) {
                    this.closeSidebarPane(frame);
                    return;
                }
                this.setSidebarPaneWidth(frame, nextWidth);
            };
            const stop = () => {
                window.removeEventListener('pointermove', move);
                window.removeEventListener('pointerup', stop);
                window.removeEventListener('pointercancel', stop);
                document.body?.classList.remove('ach-sidebar-resizing');
                if (this._stopSidebarResize === stop)
                    this._stopSidebarResize = null;
            };
            this._stopSidebarResize = stop;
            window.addEventListener('pointermove', move);
            window.addEventListener('pointerup', stop);
            window.addEventListener('pointercancel', stop);
        },
"""

__all__ = ['SIDEBAR_JS']


"""Shared data models for the visual-logic function catalog."""

from __future__ import annotations

import dataclasses


@dataclasses.dataclass(frozen=True)
class FuncEntry:
    """Metadata for a single callable in the function catalog."""

    path: str
    label: str
    signature: str
    min_args: int
    max_args: int | None
    param_names: tuple[str, ...]
    param_types: tuple[str, ...] = ()
    return_type: str = 'any'
    is_class: bool = False


@dataclasses.dataclass(frozen=True)
class TraceParam:
    """Metadata for one parameter of a plotly trace."""

    name: str
    label: str
    required: bool
    hint: str = ''


@dataclasses.dataclass(frozen=True)
class PlotlyTraceEntry:
    """Metadata for a go.<Trace> type."""

    kind: str
    label: str
    go_class: str
    params: tuple[TraceParam, ...]


def param_names_for(entry: FuncEntry) -> list[str]:
    """Return ordered parameter names for a FuncEntry."""
    return list(entry.param_names)


def param_types_for(entry: FuncEntry) -> list[str]:
    """Return ordered parameter type tags for a FuncEntry."""
    return list(entry.param_types)


"""Pure helper utilities for visual-logic code generation."""

from __future__ import annotations

import ast
from collections.abc import Callable
import re
from typing import Any


_MISSING = object()
_ConstantLiteralBuilder = Callable[[dict[str, Any]], str]


def _source_expr(
    source_id: str | None,
    node_vars: dict[str, str],
    *,
    fallback: str = 'None',
) -> str:
    """Resolve a node reference to a generated Python expression."""
    if not source_id:
        return fallback
    return node_vars.get(str(source_id), fallback)


def _literal_expr(value: Any, *, fallback: str = 'None') -> str:
    """Return a safe Python literal expression for one stored value."""
    if value is _MISSING or value in (None, ''):
        return fallback
    if isinstance(value, bool):
        return 'True' if value else 'False'
    return repr(value)


def _pin_literal_expr(
    params: dict[str, Any],
    pin_id: str,
    *,
    fallback: str = 'None',
) -> str:
    """Return one stored literal fallback expression for an input pin."""
    raw_literals = params.get('pin_literals')
    if not isinstance(raw_literals, dict):
        return fallback
    return _literal_expr(raw_literals.get(pin_id, _MISSING), fallback=fallback)


def _param_source_id(
    params: dict[str, Any],
    pin_id: str,
) -> str:
    """Return stored source id for one input pin."""
    if pin_id.startswith('arg:'):
        arg_index = int(pin_id.split(':', 1)[1])
        arg_sources = list(params.get('arg_sources') or [])
        if arg_index >= len(arg_sources):
            return ''
        return str(arg_sources[arg_index] or '')
    if pin_id.startswith('named:'):
        param_name = pin_id.split(':', 1)[1]
        named_sources = dict(params.get('named_sources') or {})
        return str(named_sources.get(param_name) or '')
    return str(params.get(pin_id) or '')


def _input_param_expr(
    params: dict[str, Any],
    pin_id: str,
    node_vars: dict[str, str],
    *,
    fallback: str = 'None',
) -> str:
    """Return connected source expression with stored literal fallback."""
    source_id = _param_source_id(params, pin_id)
    literal_expr = _pin_literal_expr(params, pin_id, fallback=fallback)
    return _source_expr(source_id, node_vars, fallback=literal_expr)


def _arg_exprs(
    params: dict[str, Any],
    node_vars: dict[str, str],
    *,
    arg_count: int | None = None,
) -> list[str]:
    """Return positional arg expressions with stored literal fallbacks."""
    arg_sources = list(params.get('arg_sources') or [])
    max_count = len(arg_sources)
    raw_literals = params.get('pin_literals')
    if isinstance(raw_literals, dict):
        for pin_id in raw_literals:
            text = str(pin_id)
            if not text.startswith('arg:'):
                continue
            try:
                arg_index = int(text.split(':', 1)[1])
            except (TypeError, ValueError):
                continue
            max_count = max(max_count, arg_index + 1)
    if arg_count is not None:
        max_count = max(max_count, arg_count)
    if arg_count is None:
        while max_count > 0:
            pin_id = f'arg:{max_count - 1}'
            source_id = _param_source_id(params, pin_id)
            literal_expr = _pin_literal_expr(params, pin_id, fallback='')
            if source_id or literal_expr:
                break
            max_count -= 1
    return [
        _input_param_expr(
            params,
            f'arg:{index}',
            node_vars,
        )
        for index in range(max_count)
    ]


def _text_constant_literal(params: dict[str, Any]) -> str:
    return repr(str(params.get('text_value') or ''))


def _bool_constant_literal(params: dict[str, Any]) -> str:
    return 'True' if bool(params.get('bool_value')) else 'False'


def _int_constant_literal(params: dict[str, Any]) -> str:
    number_value = params.get('number_value', 0)
    try:
        return repr(int(float(number_value)))
    except (TypeError, ValueError):
        return '0'


def _dict_constant_literal(params: dict[str, Any]) -> str:
    raw = str(params.get('dict_value') or '{}').strip()
    if not raw:
        return '{}'
    try:
        parsed = ast.literal_eval(raw)
    except (SyntaxError, ValueError):
        return '{}'
    return repr(parsed) if isinstance(parsed, dict) else '{}'


def _number_constant_literal(params: dict[str, Any]) -> str:
    number_value = params.get('number_value', 0.0)
    try:
        return repr(float(number_value))
    except (TypeError, ValueError):
        return '0.0'


_CONSTANT_LITERAL_BUILDERS: dict[str, _ConstantLiteralBuilder] = {
    'text': _text_constant_literal,
    'bool': _bool_constant_literal,
    'int': _int_constant_literal,
    'dict': _dict_constant_literal,
}


def _constant_literal(params: dict[str, Any]) -> str:
    """Return the Python literal expression for a constant node."""
    value_type = str(params.get('value_type') or 'number')
    builder = _CONSTANT_LITERAL_BUILDERS.get(
        value_type,
        _number_constant_literal,
    )
    return builder(params)


def _optional_index_text(value: Any) -> str:
    """Return a safe integer literal string or empty for blank values."""
    if value in (None, ''):
        return ''
    try:
        return str(int(value))
    except (TypeError, ValueError):
        return ''


def _optional_positive_int_text(value: Any) -> str:
    """Return a positive integer literal string or empty for blank values."""
    text = _optional_index_text(value)
    if not text:
        return ''
    return text if int(text) > 0 else ''


def _list_index_expr(source_expr: str, params: dict[str, Any]) -> str:
    """Return Python list or ndarray indexing or slicing expression."""
    mode = str(params.get('mode') or 'index').strip()
    if mode != 'slice':
        index = _optional_index_text(params.get('index'))
        return f'{source_expr}[{index or "0"}]'

    start = _optional_index_text(params.get('start'))
    stop = _optional_index_text(params.get('stop'))
    step = _optional_index_text(params.get('step'))
    if step == '0':
        step = ''
    slice_expr = f'{start}:{stop}'
    if step:
        slice_expr += f':{step}'
    return f'{source_expr}[{slice_expr}]'


def _safe_function_name(text: str, fallback: str) -> str:
    """Convert UI labels to valid helper function names."""
    slug = re.sub(r'[^0-9a-zA-Z_]+', '_', text.strip().lower())
    slug = re.sub(r'_+', '_', slug).strip('_')
    if not slug:
        slug = fallback
    if slug[0].isdigit():
        slug = f'{fallback}_{slug}'
    return slug


def _add_missing_pass(lines: list[str]) -> list[str]:
    """Insert pass after any block header that has no indented body."""
    result: list[str] = []
    for i, line in enumerate(lines):
        result.append(line)
        if not line.rstrip().endswith(':') or not line.strip():
            continue
        j = i + 1
        while j < len(lines) and not lines[j].strip():
            j += 1
        header_indent = len(line) - len(line.lstrip())
        if j >= len(lines):
            result.append(' ' * (header_indent + 4) + 'pass')
        else:
            next_indent = len(lines[j]) - len(lines[j].lstrip())
            if next_indent <= header_indent:
                result.append(' ' * (header_indent + 4) + 'pass')
    return result


"""Standalone host adapter and compiler for generic Acherion."""

from __future__ import annotations

from collections.abc import Callable
from typing import Any

from acherion.catalog import modules as _catalog_modules
from acherion.compiler.orchestration import (
    _GraphEventCompilerBase,
    _GraphNodeEmitterBase,
    _GraphRunCompilerBase,
)
from acherion.events import (
    AcherionExternalEvent,
    EXTERNAL_EVENT_NODE_KIND,
    RUN_EVENT_KEY,
    acherion_external_event_params,
    collect_acherion_external_events,
    default_acherion_external_events,
    serializable_acherion_external_events,
)
from acherion.model import AcherionGraph, AcherionNode, _system_node_id
from acherion.preview import AcherionPreviewRunResult


_RUNTIME_NORMALIZER_SIGNATURE = 'def _normalize_scoped_bindings(bindings=None):'
_RUNTIME_BINDINGS_LINE = 'bindings = _normalize_scoped_bindings(bindings)'
_RUNTIME_LOCALS_RETURN = 'return dict(locals())'


def _build_system_node(
    *,
    existing: AcherionNode | None,
    kind: str,
    key: str,
    title: str,
    system_group: str,
    params: dict[str, Any],
) -> AcherionNode:
    """Merge one system-node definition onto an existing node."""
    node = existing or AcherionNode(
        node_id=_system_node_id(kind, key),
        kind=kind,
    )
    merged = dict(node.params)
    merged.update(params)
    merged['system_node'] = True
    merged['system_group'] = system_group
    node.node_id = _system_node_id(kind, key)
    node.kind = kind
    node.title = title
    node.params = merged
    return node


class _StandaloneGraphCompilerBase(_GraphNodeEmitterBase):
    """Standalone-specific node-emission hooks."""

    def _emit_host_fallback_node(
        self,
        *,
        node: AcherionNode,
        params: dict[str, object],
        var_name: str,
        indent: str,
    ) -> None:
        del params, var_name
        self._state.lines.append(
            f'{indent}pass  # Unsupported standalone node kind: {node.kind}'
        )


class _StandaloneGraphRunCompiler(
    _GraphRunCompilerBase,
    _StandaloneGraphCompilerBase,
):
    """Compile a generic standalone graph into runnable Python code."""

    def _append_run_method(self) -> None:
        self._state.lines.extend([
            'def run(bindings=None):',
            '    bindings = _normalize_scoped_bindings(bindings)',
            '',
        ])
        manual_nodes = [
            node
            for node in self._graph.nodes
            if not bool(node.params.get('system_node'))
        ]
        if not manual_nodes:
            self._state.lines.append('    return dict(locals())')
            return
        self._scope.emit(self._emit_graph_node)
        self._append_passive_value_roots()
        self._state.lines.append('    return dict(locals())')

    def _append_passive_value_roots(self) -> None:
        for node in self._top_level_nodes:
            if bool(node.params.get('system_node')):
                continue
            self._scope.ensure_source(
                node.node_id,
                indent_level=1,
                emit_node=self._emit_graph_node,
            )

    def _create_event_compiler(
        self,
        event_node: AcherionNode,
    ) -> '_StandaloneGraphEventCompiler':
        return _StandaloneGraphEventCompiler(
            self._graph,
            function_graph=self._function_graph,
            original_index=self._original_index,
            event_node=event_node,
        )


class _StandaloneGraphEventCompiler(
    _GraphEventCompilerBase,
    _StandaloneGraphCompilerBase,
):
    """Compile one non-run external event handler."""

    def _handler_prelude_lines(self) -> list[str]:
        return [
            f'def {self._handler_name}(bindings=None):',
            '    bindings = _normalize_scoped_bindings(bindings)',
            '',
        ]


def _standalone_external_events(
    graph: AcherionGraph,
) -> dict[str, dict[str, str]]:
    """Return generated external-event metadata for one graph."""
    return serializable_acherion_external_events(
        collect_acherion_external_events(graph.nodes)
    )


def _build_standalone_runtime_source(
    *,
    compiled_body: str,
    external_events: dict[str, dict[str, str]],
) -> str:
    """Return full standalone runtime source from compiled graph body."""
    lines = [
        'from __future__ import annotations',
        '',
        'import collections',
        'import logging',
        'import pathlib',
        'import re',
        'import typing',
        '',
        'logger = logging.getLogger(__name__)',
        '',
        f'ACHERION_EXTERNAL_EVENTS = {external_events!r}',
        '',
        'def _normalize_scoped_bindings(bindings=None):',
        '    normalized = {}',
        '    raw_bindings = dict(bindings or {})',
        '    for scope, values in raw_bindings.items():',
        "        clean_scope = str(scope or '').strip()",
        '        if not clean_scope:',
        '            continue',
        '        if isinstance(values, collections.abc.Mapping):',
        '            normalized[clean_scope] = dict(values)',
        '        else:',
        '            normalized[clean_scope] = {}',
        '    return normalized',
        '',
        compiled_body,
    ]
    return '\n'.join(lines).rstrip() + '\n'


def _compile_standalone_runtime(
    graph: AcherionGraph,
) -> tuple[str, dict[str, str], dict[str, dict[str, str]]]:
    """Compile one graph and return source plus preview mapping metadata."""
    external_events = _standalone_external_events(graph)
    compiler = _StandaloneGraphRunCompiler(graph)
    compiled_body = compiler.compile().rstrip()
    source_code = _build_standalone_runtime_source(
        compiled_body=compiled_body,
        external_events=external_events,
    )
    return source_code, compiler.node_var_map(), external_events


def compile_acherion_graph(graph: AcherionGraph) -> str:
    """Compile one Acherion graph into standalone Python source."""
    source_code, _node_var_map, _external_events = _compile_standalone_runtime(
        graph
    )
    del _node_var_map, _external_events
    return source_code


def standalone_editor_visible_code(source_code: str) -> str:
    """Hide generated standalone runtime scaffold in the code pane."""
    lines = str(source_code or '').splitlines()
    normalizer_index = -1
    for index, line in enumerate(lines):
        if line.startswith(_RUNTIME_NORMALIZER_SIGNATURE):
            normalizer_index = index
            break
    if normalizer_index >= 0:
        end_index = normalizer_index + 1
        while end_index < len(lines):
            line = lines[end_index]
            if line and not line.startswith('    '):
                break
            end_index += 1
        lines = lines[end_index:]

    cleaned_lines: list[str] = []
    previous_blank = False
    for line in lines:
        stripped = line.strip()
        if stripped in {
            _RUNTIME_BINDINGS_LINE,
            _RUNTIME_LOCALS_RETURN,
        }:
            continue
        if stripped.startswith('ACHERION_EXTERNAL_EVENTS = '):
            continue
        if '(bindings=None):' in line and line.startswith('def '):
            line = line.replace('(bindings=None):', '():')
        is_blank = not stripped
        if is_blank and previous_blank:
            continue
        cleaned_lines.append(line)
        previous_blank = is_blank
    return '\n'.join(cleaned_lines).strip()


def load_acherion_graph_namespace(source_code: str) -> dict[str, Any]:
    """Load generated Acherion code into a fresh Python namespace."""
    namespace: dict[str, Any] = {
        '__builtins__': __builtins__,
        '__name__': '_acherion_standalone_runtime',
        **_catalog_modules.runtime_global_bindings(),
    }
    exec(source_code, namespace)  # pylint: disable=exec-used
    return namespace


def execute_acherion_graph(
    source_code: str,
    *,
    event_key: str = RUN_EVENT_KEY,
    bindings: dict[str, dict[str, Any]] | None = None,
) -> dict[str, Any]:
    """Execute one generated Acherion handler and return runtime locals."""
    namespace = load_acherion_graph_namespace(source_code)
    event_map = dict(namespace.get('ACHERION_EXTERNAL_EVENTS') or {})
    default_handler_name = 'run'
    if event_key != RUN_EVENT_KEY:
        default_handler_name = acherion_event_handler_name(event_key)
    event_meta = dict(event_map.get(event_key) or {})
    handler_name = str(
        event_meta.get('handler_name') or default_handler_name
    ).strip()
    handler = namespace.get(handler_name)
    if not callable(handler):
        raise ValueError(
            f'Generated graph does not define a handler for {event_key!r}.'
        )
    result = handler(bindings=bindings)
    if isinstance(result, dict):
        return result
    return {'result': result}


def run_standalone_acherion_preview(
    graph: AcherionGraph,
    *,
    bindings: dict[str, dict[str, Any]] | None = None,
    event_key: str = RUN_EVENT_KEY,
) -> AcherionPreviewRunResult:
    """Execute one standalone graph preview and map values back to nodes."""
    source_code, node_var_map, external_events = _compile_standalone_runtime(
        graph
    )
    local_values = execute_acherion_graph(
        source_code,
        event_key=event_key,
        bindings=bindings,
    )
    reference_values = {
        source_id: local_values[var_name]
        for source_id, var_name in node_var_map.items()
        if var_name in local_values
    }
    event_meta = dict(external_events.get(event_key) or {})
    event_title = str(event_meta.get('title') or event_key).strip()
    if event_title.startswith('Event: '):
        event_title = event_title[7:]
    hint = f'Preview ran for {event_title or event_key}.'
    if not reference_values:
        hint += ' No previewable values were produced.'
    return AcherionPreviewRunResult(
        reference_values=reference_values,
        state_values={
            'hint': hint,
            'acherion.preview_event': event_key,
            'acherion.preview_value_count': len(reference_values),
        },
    )


class StandaloneAcherionHost:
    """Generic host used by the standalone Acherion app."""

    def external_events(self) -> dict[str, AcherionExternalEvent]:
        """Return standalone host events available inside the designer."""
        return default_acherion_external_events()

    def generated_user_code(self, designer: Any) -> str:
        """Compile the current graph into standalone Python source."""
        return compile_acherion_graph(designer.graph_state())

    def generated_runtime_bindings(
        self,
        designer: Any,
    ) -> dict[str, dict[str, dict[str, Any]]]:
        """Return standalone runtime bindings for generated code."""
        del designer
        return {}

    def sync_manual_schema_keys(self, designer: Any) -> None:
        """No-op for the standalone host."""
        del designer

    def sync_system_nodes(self, designer: Any) -> None:
        """Materialize generic external-event source nodes."""
        manual_nodes = designer._manual_nodes()
        consumed_ids = {node.node_id for node in manual_nodes}
        source_nodes: list[AcherionNode] = []
        for index, event in enumerate(designer._external_events()):
            existing = designer._take_matching_node(
                consumed_ids=consumed_ids,
                kind=EXTERNAL_EVENT_NODE_KIND,
                key=event.event_key,
                param_key='event_key',
            )
            node = _build_system_node(
                existing=existing,
                kind=EXTERNAL_EVENT_NODE_KIND,
                key=event.event_key,
                title=event.title,
                system_group='source',
                params=acherion_external_event_params(event),
            )
            designer._seed_position(node, group='source', index=index)
            source_nodes.append(node)
        designer._graph.nodes = [*source_nodes, *manual_nodes]
        designer._ensure_function_box_entries()
        designer._sync_function_box_ports()

    def render_system_sink_config_fields(
        self,
        designer: Any,
        node: Any,
        *,
        refresh_editor: Callable[[], None] | None,
        apply_change: Callable[..., None],
    ) -> bool:
        """Standalone host defines no system sink config fields."""
        del designer, node, refresh_editor, apply_change
        return False

    def render_node_config_fields(
        self,
        designer: Any,
        node: Any,
        *,
        refresh_editor: Callable[[], None] | None,
        apply_change: Callable[..., None],
    ) -> bool:
        """Standalone host owns no extra node config fields."""
        del designer, node, refresh_editor, apply_change
        return False


__all__ = [
    'StandaloneAcherionHost',
    'compile_acherion_graph',
    'execute_acherion_graph',
    'load_acherion_graph_namespace',
    'run_standalone_acherion_preview',
    'standalone_editor_visible_code',
]



"""CSS fragments for Acherion asset bundles."""

from __future__ import annotations

from acherion.assets.css.chrome import CHROME_CSS
from acherion.assets.css.nodes import NODE_CSS
from acherion.assets.css.search import SEARCH_CSS
from acherion.assets.css.workbench import WORKBENCH_CSS

__all__ = ['CHROME_CSS', 'NODE_CSS', 'SEARCH_CSS', 'WORKBENCH_CSS']


"""Core source node definitions for Acherion."""

from __future__ import annotations

from acherion.events import EXTERNAL_EVENT_NODE_KIND
import acherion.node as acherion_node


class ExternalEventNode(acherion_node.SystemSourceNodeDefinition):
    kind = EXTERNAL_EVENT_NODE_KIND
    label = 'External Event'
    icon = 'bolt'
    tooltip = 'Fire logic from a host-driven or component-driven runtime event.'
    flavor = 'event'
    exec_out = True


BUILTIN_SOURCE_NODES = (
    ExternalEventNode(),
)

__all__ = [
    'BUILTIN_SOURCE_NODES',
]



"""Theme helpers for standalone and embedded Acherion workbenches."""

from __future__ import annotations

import importlib.resources
import re
from collections.abc import Mapping
from functools import lru_cache
from typing import Any


AcherionThemeOverrides = Mapping[str, str]

_THEME_RESOURCE = 'theme.css'

_DEFAULT_THEME_VARIABLES: dict[str, str] = {
    '--oe-bg': '#000000',
    '--oe-panel': '#000000',
    '--oe-border': '#2F3336',
    '--oe-surface': '#000000',
    '--oe-text': '#E7E9EA',
    '--oe-muted': '#71767B',
    '--oe-blue': '#1D9BF0',
    '--oe-green': '#00BA7C',
    '--oe-red': '#F4212E',
    '--oe-yellow': '#FFD400',
    '--oe-hover-tint': 'rgba(255, 255, 255, 0.07)',
    '--oe-hover-tint-subtle': 'rgba(255, 255, 255, 0.03)',
    '--oe-hover-border': 'rgba(255, 255, 255, 0.4)',
    '--oe-input-hover-border': 'rgba(255, 255, 255, 0.4)',
    '--oe-uploader-hover-border': 'rgba(255, 255, 255, 0.35)',
    '--oe-header-bg': 'rgba(0, 0, 0, 0.85)',
    '--oe-dialog-shadow': '0 8px 40px rgba(0, 0, 0, 0.9)',
    '--oe-legend-bgcolor': 'rgba(22, 24, 28, 0.85)',
    '--oe-uploader-header-bg': 'rgba(255, 255, 255, 0.04)',
    '--oe-inverted-bg': '#E7E9EA',
    '--oe-inverted-text': '#0F1419',
    '--oe-color-scheme': 'dark',
    '--oe-menu-shadow': '0 8px 32px rgba(0, 0, 0, 0.85)',
    '--oe-date-indicator-filter': 'brightness(0) invert(1)',
    '--oe-spin-button-filter': 'invert(1)',
    '--ach-sidebar-glass-bg': 'rgba(0,0,0,0.72)',
    '--ach-sidebar-glass-filter': 'blur(16px) saturate(112%)',
    '--ach-sidebar-panel-bg': 'rgba(255,255,255,0.03)',
    '--ach-sidebar-panel-bg-strong': 'rgba(255,255,255,0.04)',
}

_THEME_ALIAS_TO_VARIABLE: dict[str, str] = {
    'bg': '--oe-bg',
    'panel': '--oe-panel',
    'border': '--oe-border',
    'surface': '--oe-surface',
    'text': '--oe-text',
    'muted': '--oe-muted',
    'blue': '--oe-blue',
    'green': '--oe-green',
    'red': '--oe-red',
    'yellow': '--oe-yellow',
    'primary': '--oe-blue',
    'secondary': '--oe-muted',
    'accent': '--oe-blue',
    'positive': '--oe-green',
    'negative': '--oe-red',
    'warning': '--oe-yellow',
    'info': '--oe-blue',
    'hover_tint': '--oe-hover-tint',
    'hover_tint_subtle': '--oe-hover-tint-subtle',
    'hover_border': '--oe-hover-border',
    'input_hover_border': '--oe-input-hover-border',
    'uploader_hover_border': '--oe-uploader-hover-border',
    'header_bg': '--oe-header-bg',
    'dialog_shadow': '--oe-dialog-shadow',
    'legend_bgcolor': '--oe-legend-bgcolor',
    'uploader_header_bg': '--oe-uploader-header-bg',
    'inverted_bg': '--oe-inverted-bg',
    'inverted_text': '--oe-inverted-text',
    'color_scheme': '--oe-color-scheme',
    'menu_shadow': '--oe-menu-shadow',
    'date_indicator_filter': '--oe-date-indicator-filter',
    'spin_button_filter': '--oe-spin-button-filter',
    'sidebar_glass_bg': '--ach-sidebar-glass-bg',
    'sidebar_glass_filter': '--ach-sidebar-glass-filter',
    'sidebar_panel_bg': '--ach-sidebar-panel-bg',
    'sidebar_panel_bg_strong': '--ach-sidebar-panel-bg-strong',
}

_DEFAULT_PLOTLY_FONT_SIZE = 15
_PLOTLY_FONT_FAMILY = (
    '-apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Helvetica, '
    'Arial, sans-serif'
)
_PLOTLY_COLORWAY = ('--oe-blue', '--oe-green', '--oe-red', '--oe-yellow')
_PLOTLY_EXTRA_COLORWAY = ('#FF7A00', '--oe-muted', '#7856FF')

_COMMENT_RE = re.compile(r'/\*.*?\*/', flags=re.S)
_RULE_RE = re.compile(r'(?s)([^{}]+)\{([^{}]*)\}')


@lru_cache(maxsize=1)
def _load_default_theme_css() -> str:
    """Return packaged Acherion theme CSS."""
    return importlib.resources.files('acherion').joinpath(
        _THEME_RESOURCE
    ).read_text(encoding='utf-8')


def _resolve_theme_variable_name(key: str) -> str:
    """Normalize one public theme key into a CSS custom-property name."""
    clean_key = str(key or '').strip()
    if not clean_key:
        raise ValueError('Theme override keys must not be empty.')
    if clean_key.startswith('--'):
        return clean_key
    variable_name = _THEME_ALIAS_TO_VARIABLE.get(clean_key)
    if variable_name is None:
        raise ValueError(
            'Unknown theme override key: '
            f'{clean_key!r}. Use a supported alias or a raw CSS variable name.'
        )
    return variable_name


def normalize_acherion_theme_overrides(
    theme_overrides: AcherionThemeOverrides | None,
) -> dict[str, str]:
    """Return normalized CSS custom-property overrides for Acherion."""
    normalized: dict[str, str] = {}
    for key, value in dict(theme_overrides or {}).items():
        clean_value = str(value or '').strip()
        if not clean_value:
            continue
        normalized[_resolve_theme_variable_name(key)] = clean_value
    return normalized


def acherion_ui_colors(
    theme_overrides: AcherionThemeOverrides | None = None,
) -> dict[str, str]:
    """Return NiceGUI/Quasar color tokens derived from Acherion theme values."""
    theme_values = dict(_DEFAULT_THEME_VARIABLES)
    theme_values.update(normalize_acherion_theme_overrides(theme_overrides))
    return {
        'primary': theme_values['--oe-blue'],
        'secondary': theme_values['--oe-muted'],
        'accent': theme_values['--oe-blue'],
        'positive': theme_values['--oe-green'],
        'negative': theme_values['--oe-red'],
        'warning': theme_values['--oe-yellow'],
        'info': theme_values['--oe-blue'],
    }


def acherion_plotly_template_payload(
    theme_overrides: AcherionThemeOverrides | None = None,
) -> dict[str, Any]:
    """Return a serialized Plotly template payload for Acherion's dark theme."""
    theme_values = dict(_DEFAULT_THEME_VARIABLES)
    theme_values.update(normalize_acherion_theme_overrides(theme_overrides))
    colorway = [
        theme_values[color_key]
        for color_key in _PLOTLY_COLORWAY
        if color_key in theme_values
    ]
    colorway.extend(_PLOTLY_EXTRA_COLORWAY)
    base_font = {
        'color': theme_values['--oe-text'],
        'family': _PLOTLY_FONT_FAMILY,
        'size': _DEFAULT_PLOTLY_FONT_SIZE,
    }
    muted_font = {
        'color': theme_values['--oe-muted'],
        'size': _DEFAULT_PLOTLY_FONT_SIZE - 2,
    }
    title_font = {
        'color': theme_values['--oe-text'],
        'size': _DEFAULT_PLOTLY_FONT_SIZE + 2,
    }
    axis_title_font = {
        'color': theme_values['--oe-text'],
        'size': _DEFAULT_PLOTLY_FONT_SIZE - 1,
    }
    return {
        'layout': {
            'paper_bgcolor': 'rgba(0,0,0,0)',
            'plot_bgcolor': 'rgba(0,0,0,0)',
            'font': base_font,
            'title': {'font': title_font},
            'xaxis': {
                'gridcolor': theme_values['--oe-border'],
                'zerolinecolor': theme_values['--oe-border'],
                'linecolor': theme_values['--oe-border'],
                'tickcolor': theme_values['--oe-muted'],
                'tickfont': muted_font,
                'title': {'font': axis_title_font},
            },
            'yaxis': {
                'gridcolor': theme_values['--oe-border'],
                'zerolinecolor': theme_values['--oe-border'],
                'linecolor': theme_values['--oe-border'],
                'tickcolor': theme_values['--oe-muted'],
                'tickfont': muted_font,
                'title': {'font': axis_title_font},
            },
            'polar': {
                'bgcolor': 'rgba(0,0,0,0)',
                'angularaxis': {
                    'gridcolor': theme_values['--oe-border'],
                    'linecolor': theme_values['--oe-border'],
                    'tickcolor': theme_values['--oe-muted'],
                    'tickfont': muted_font,
                },
                'radialaxis': {
                    'gridcolor': theme_values['--oe-border'],
                    'linecolor': theme_values['--oe-border'],
                    'tickcolor': theme_values['--oe-muted'],
                    'tickfont': muted_font,
                    'title': {'font': axis_title_font},
                },
            },
            'legend': {
                'bgcolor': theme_values['--oe-legend-bgcolor'],
                'bordercolor': theme_values['--oe-border'],
                'borderwidth': 1,
                'font': {
                    'color': theme_values['--oe-text'],
                    'size': _DEFAULT_PLOTLY_FONT_SIZE - 1,
                },
            },
            'colorway': colorway,
            'margin': {'l': 48, 'r': 24, 't': 40, 'b': 40},
            'hoverlabel': {
                'bgcolor': theme_values['--oe-surface'],
                'bordercolor': theme_values['--oe-border'],
                'font': {'color': theme_values['--oe-text']},
            },
        },
        'data': {
            'heatmap': [
                {
                    'hovertemplate': (
                        'X: %{x:.3f}<br>Y: %{y:.3f}<br>Value: %{z}'
                        '<extra></extra>'
                    ),
                }
            ],
            'scatter': [
                {
                    'hovertemplate': (
                        'X: %{x:.3f}<br>Y: %{y:.3f}<br>%{fullData.name}'
                        '<extra></extra>'
                    ),
                }
            ],
        },
    }


def build_acherion_theme_override_css(
    *,
    scope_selector: str,
    theme_overrides: AcherionThemeOverrides | None = None,
) -> str:
    """Build CSS custom-property overrides for one Acherion scope."""
    clean_scope_selector = str(scope_selector or '').strip()
    if not clean_scope_selector:
        raise ValueError('A scope selector is required for theme overrides.')
    normalized = normalize_acherion_theme_overrides(theme_overrides)
    if not normalized:
        return ''
    lines = [f'{clean_scope_selector} {{']
    lines.extend(
        f'    {variable_name}: {value};'
        for variable_name, value in normalized.items()
    )
    lines.append('}')
    return '\n'.join(lines)


def _scope_selector(selector: str, scope_selector: str) -> tuple[str, ...]:
    """Return one CSS selector rewritten to target an Acherion subtree."""
    clean_selector = str(selector or '').strip()
    if not clean_selector:
        return ()
    if clean_selector == ':root':
        return (scope_selector,)
    if clean_selector in {'body', '.q-app', '.q-page', '.nicegui-content'}:
        return (scope_selector,)
    if clean_selector == '*':
        return (scope_selector, f'{scope_selector} *')
    if clean_selector == '*::before':
        return (f'{scope_selector}::before', f'{scope_selector} *::before')
    if clean_selector == '*::after':
        return (f'{scope_selector}::after', f'{scope_selector} *::after')
    return (f'{scope_selector} {clean_selector}',)


@lru_cache(maxsize=64)
def build_scoped_acherion_theme_css(scope_selector: str) -> str:
    """Return the packaged Acherion theme CSS scoped to one container."""
    clean_scope_selector = str(scope_selector or '').strip()
    if not clean_scope_selector:
        raise ValueError('A scope selector is required for scoped theme CSS.')
    if clean_scope_selector == ':root':
        return _load_default_theme_css()

    raw_css = _COMMENT_RE.sub('', _load_default_theme_css())
    scoped_rules: list[str] = []
    for match in _RULE_RE.finditer(raw_css):
        selectors = [
            selector.strip()
            for selector in str(match.group(1) or '').split(',')
            if selector.strip()
        ]
        declarations = str(match.group(2) or '').strip()
        if not selectors or not declarations:
            continue
        scoped_selectors: list[str] = []
        seen_selectors: set[str] = set()
        for selector in selectors:
            for scoped_selector in _scope_selector(selector, clean_scope_selector):
                if scoped_selector in seen_selectors:
                    continue
                seen_selectors.add(scoped_selector)
                scoped_selectors.append(scoped_selector)
        if not scoped_selectors:
            continue
        scoped_rules.append(
            ',\n'.join(scoped_selectors) + ' {\n' + declarations + '\n}'
        )
    return '\n\n'.join(scoped_rules)


def build_embedded_acherion_theme_css(
    *,
    scope_selector: str,
    theme_overrides: AcherionThemeOverrides | None = None,
) -> str:
    """Return fully scoped Acherion theme CSS for one embedded workbench."""
    base_css = build_scoped_acherion_theme_css(scope_selector)
    override_css = build_acherion_theme_override_css(
        scope_selector=scope_selector,
        theme_overrides=theme_overrides,
    )
    if not override_css:
        return base_css
    return base_css + '\n\n' + override_css


__all__ = [
    'AcherionThemeOverrides',
    'acherion_plotly_template_payload',
    'acherion_ui_colors',
    'build_acherion_theme_override_css',
    'build_embedded_acherion_theme_css',
    'build_scoped_acherion_theme_css',
    'normalize_acherion_theme_overrides',
]


"""Shared node-emission helpers for visual-logic compilation."""

from __future__ import annotations

from collections.abc import Callable
from typing import Any

from acherion.catalog import plotly as _catalog_plotly
from acherion.compiler.graph import (
    _FunctionBoxGraphView,
)
from acherion.compiler.utils import (
    _arg_exprs,
    _constant_literal,
    _input_param_expr,
    _list_index_expr,
    _safe_function_name,
)
from acherion.compiler.state import EmitState
from acherion.model import AcherionNode


_EmitNodeHandler = Callable[..., bool]
_UNARY_EXPR_TEMPLATES: dict[str, str] = {
    'negate': '-({source})',
    'round': 'round({source})',
}


def _emit_function_box_node(
    *,
    state: EmitState,
    node: AcherionNode,
    params: dict[str, Any],
    function_graph: _FunctionBoxGraphView,
    indent: str,
    var_name: str,
    runtime_self_name: str | None = 'self',
    helper_as_attribute: bool = True,
) -> bool:
    """Emit code for a function_box node when present."""
    if node.kind != 'function_box':
        return False

    helper_name = _safe_function_name(
        str(params.get('function_name') or node.title or node.node_id),
        f'function_{node.node_id}',
    )
    boundary_inputs = function_graph.boundary_input_sources(
        node.node_id,
    )
    args = [
        state.source_expr(source_id)
        for source_id in boundary_inputs
    ]
    return_source_ids = function_graph.boundary_output_sources(
        node.node_id,
    )
    call_args = list(args)
    if not helper_as_attribute and runtime_self_name is not None:
        call_args = [runtime_self_name, *call_args]
    call_target = (
        f'{runtime_self_name}.{helper_name}'
        if helper_as_attribute else helper_name
    )
    call_expr = f"{call_target}({', '.join(call_args)})"
    if not return_source_ids:
        state.lines.append(f'{indent}{call_expr}')
        return True
    if len(return_source_ids) == 1:
        state.lines.append(f'{indent}{var_name} = {call_expr}')
        state.store_source(return_source_ids[0], var_name)
        return True
    tuple_var = f'{var_name}_tuple'
    state.lines.append(f'{indent}{tuple_var} = {call_expr}')
    for out_index, source_id in enumerate(return_source_ids):
        out_var = f'{var_name}_{out_index + 1}'
        state.lines.append(f'{indent}{out_var} = {tuple_var}[{out_index}]')
        state.store_source(source_id, out_var)
    return True


def _emit_constant_node(
    *,
    state: EmitState,
    node: AcherionNode,
    params: dict[str, Any],
    indent: str,
    var_name: str,
    method_owner_name: str | None = 'self',
) -> bool:
    del method_owner_name
    state.lines.append(f'{indent}{var_name} = {_constant_literal(params)}')
    state.store(node.node_id, var_name)
    return True


def _emit_callable_node(
    *,
    state: EmitState,
    node: AcherionNode,
    params: dict[str, Any],
    indent: str,
    var_name: str,
    method_owner_name: str | None = 'self',
) -> bool:
    function_path = str(params.get('function_path') or '').strip()
    call_target = function_path
    if function_path.startswith('user.'):
        helper_name = function_path.split('.', 1)[1]
        call_target = (
            f'{method_owner_name}.{helper_name}'
            if method_owner_name else helper_name
        )
    args = _arg_exprs(params, state.node_vars)
    result_var_name = var_name
    if call_target and '.' not in call_target and call_target == var_name:
        result_var_name = f'{var_name}_result'
    if call_target:
        state.lines.append(
            f"{indent}{result_var_name} = {call_target}({', '.join(args)})"
        )
    else:
        state.lines.append(
            f'{indent}{result_var_name} = None  # Set a function path.'
        )
    state.store(node.node_id, result_var_name)
    return True


def _emit_call_method_node(
    *,
    state: EmitState,
    node: AcherionNode,
    params: dict[str, Any],
    indent: str,
    var_name: str,
    method_owner_name: str | None = 'self',
) -> bool:
    del method_owner_name
    instance_id = str(params.get('instance') or '').strip()
    method_name = str(params.get('method_name') or '').strip()
    instance_expr = state.source_expr(instance_id)
    args = _arg_exprs(params, state.node_vars)
    if instance_id and method_name:
        state.lines.append(
            f"{indent}{var_name} = "
            f"{instance_expr}.{method_name}({', '.join(args)})"
        )
    else:
        state.lines.append(
            f'{indent}{var_name} = None'
            f'  # Connect instance and set method name.'
        )
    state.store(node.node_id, var_name)
    return True


def _emit_get_attribute_node(
    *,
    state: EmitState,
    node: AcherionNode,
    params: dict[str, Any],
    indent: str,
    var_name: str,
    method_owner_name: str | None = 'self',
) -> bool:
    del method_owner_name
    instance_id = str(params.get('instance') or '').strip()
    attr_name = str(params.get('attribute_name') or '').strip()
    instance_expr = state.source_expr(instance_id)
    if instance_id and attr_name:
        state.lines.append(
            f'{indent}{var_name} = {instance_expr}.{attr_name}'
        )
    else:
        state.lines.append(
            f'{indent}{var_name} = None'
            f'  # Connect instance and set attribute name.'
        )
    state.store(node.node_id, var_name)
    return True


def _emit_set_attribute_node(
    *,
    state: EmitState,
    node: AcherionNode,
    params: dict[str, Any],
    indent: str,
    var_name: str,
    method_owner_name: str | None = 'self',
) -> bool:
    del var_name, method_owner_name
    instance_id = str(params.get('instance') or '').strip()
    attr_name = str(params.get('attribute_name') or '').strip()
    instance_expr = state.source_expr(instance_id)
    value_expr = _input_param_expr(
        params,
        'value',
        state.node_vars,
    )
    if instance_id and attr_name:
        state.lines.append(
            f'{indent}{instance_expr}.{attr_name} = {value_expr}'
        )
    else:
        state.lines.append(
            f'{indent}pass'
            f'  # Connect instance and set attribute name.'
        )
    return True


def _emit_for_each_node(
    *,
    state: EmitState,
    node: AcherionNode,
    params: dict[str, Any],
    indent: str,
    var_name: str,
    method_owner_name: str | None = 'self',
) -> bool:
    del var_name, method_owner_name
    list_id = str(params.get('list') or '').strip()
    list_expr = state.source_expr(list_id)
    item_var = f'_item_{node.node_id}'
    idx_var = f'_idx_{node.node_id}'
    state.store(node.node_id, item_var)
    state.store(node.node_id, item_var, 0)
    state.store(node.node_id, idx_var, 1)
    state.lines.append(
        f'{indent}for {idx_var}, {item_var} in enumerate({list_expr}):'
    )
    return True


def _emit_collect_node(
    *,
    state: EmitState,
    node: AcherionNode,
    params: dict[str, Any],
    indent: str,
    var_name: str,
    method_owner_name: str | None = 'self',
) -> bool:
    del node, method_owner_name
    value_expr = _input_param_expr(
        params,
        'value',
        state.node_vars,
    )
    state.lines.append(f'{indent}{var_name}.append({value_expr})')
    return True


def _emit_compare_node(
    *,
    state: EmitState,
    node: AcherionNode,
    params: dict[str, Any],
    indent: str,
    var_name: str,
    method_owner_name: str | None = 'self',
) -> bool:
    del method_owner_name
    left_expr = _input_param_expr(
        params,
        'left_source',
        state.node_vars,
    )
    right_expr = _input_param_expr(
        params,
        'right_source',
        state.node_vars,
    )
    operator = str(params.get('operator') or '>')
    state.lines.append(
        f'{indent}{var_name} = ({left_expr} {operator} {right_expr})'
    )
    state.store(node.node_id, var_name)
    return True


def _emit_branch_value_node(
    *,
    state: EmitState,
    node: AcherionNode,
    params: dict[str, Any],
    indent: str,
    var_name: str,
    method_owner_name: str | None = 'self',
) -> bool:
    del method_owner_name
    cond = _input_param_expr(
        params,
        'condition_source',
        state.node_vars,
        fallback='False',
    )
    true_expr = _input_param_expr(
        params,
        'true_source',
        state.node_vars,
    )
    false_expr = _input_param_expr(
        params,
        'false_source',
        state.node_vars,
    )
    state.lines.append(
        f'{indent}{var_name} = '
        f'{true_expr} if bool({cond}) else {false_expr}'
    )
    state.store(node.node_id, var_name)
    return True


def _emit_op_arithmetic_node(
    *,
    state: EmitState,
    node: AcherionNode,
    params: dict[str, Any],
    indent: str,
    var_name: str,
    method_owner_name: str | None = 'self',
) -> bool:
    del method_owner_name
    operator = str(params.get('operator') or '+')
    left = _input_param_expr(
        params,
        'left_source',
        state.node_vars,
    )
    right = _input_param_expr(
        params,
        'right_source',
        state.node_vars,
    )
    state.lines.append(
        f'{indent}{var_name} = ({left}) {operator} ({right})'
    )
    state.store(node.node_id, var_name)
    return True


def _emit_op_unary_node(
    *,
    state: EmitState,
    node: AcherionNode,
    params: dict[str, Any],
    indent: str,
    var_name: str,
    method_owner_name: str | None = 'self',
) -> bool:
    del method_owner_name
    func = str(params.get('function') or 'abs')
    src = _input_param_expr(
        params,
        'source',
        state.node_vars,
    )
    template = _UNARY_EXPR_TEMPLATES.get(func, '{func}({source})')
    expr = template.format(func=func, source=src)
    state.lines.append(f'{indent}{var_name} = {expr}')
    state.store(node.node_id, var_name)
    return True


def _emit_op_logic_node(
    *,
    state: EmitState,
    node: AcherionNode,
    params: dict[str, Any],
    indent: str,
    var_name: str,
    method_owner_name: str | None = 'self',
) -> bool:
    del method_owner_name
    operator = str(params.get('operator') or 'and')
    left = _input_param_expr(
        params,
        'left_source',
        state.node_vars,
        fallback='False',
    )
    right = _input_param_expr(
        params,
        'right_source',
        state.node_vars,
        fallback='False',
    )
    state.lines.append(
        f'{indent}{var_name} = bool({left}) {operator} bool({right})'
    )
    state.store(node.node_id, var_name)
    return True


def _emit_op_not_node(
    *,
    state: EmitState,
    node: AcherionNode,
    params: dict[str, Any],
    indent: str,
    var_name: str,
    method_owner_name: str | None = 'self',
) -> bool:
    del method_owner_name
    src = _input_param_expr(
        params,
        'source',
        state.node_vars,
        fallback='False',
    )
    state.lines.append(f'{indent}{var_name} = not bool({src})')
    state.store(node.node_id, var_name)
    return True


def _emit_make_list_node(
    *,
    state: EmitState,
    node: AcherionNode,
    params: dict[str, Any],
    indent: str,
    var_name: str,
    method_owner_name: str | None = 'self',
) -> bool:
    del method_owner_name
    arg_count = max(0, int(params.get('arg_count', 0) or 0))
    args = _arg_exprs(
        params,
        state.node_vars,
        arg_count=arg_count,
    )
    state.lines.append(f"{indent}{var_name} = [{', '.join(args)}]")
    state.store(node.node_id, var_name)
    return True


def _emit_make_dict_node(
    *,
    state: EmitState,
    node: AcherionNode,
    params: dict[str, Any],
    indent: str,
    var_name: str,
    method_owner_name: str | None = 'self',
) -> bool:
    del method_owner_name
    arg_count = max(0, int(params.get('arg_count', 0) or 0))
    args = _arg_exprs(
        params,
        state.node_vars,
        arg_count=arg_count,
    )
    key_names = list(params.get('key_names') or [])
    entries = []
    for index, expr in enumerate(args):
        default_key = f'key_{index + 1}'
        key_name = str(
            key_names[index] if index < len(key_names) else default_key
        ).strip() or default_key
        entries.append(f'{key_name!r}: {expr}')
    state.lines.append(f"{indent}{var_name} = {{{', '.join(entries)}}}")
    state.store(node.node_id, var_name)
    return True


def _emit_list_index_node(
    *,
    state: EmitState,
    node: AcherionNode,
    params: dict[str, Any],
    indent: str,
    var_name: str,
    method_owner_name: str | None = 'self',
) -> bool:
    del method_owner_name
    src = _input_param_expr(
        params,
        'source',
        state.node_vars,
    )
    state.lines.append(
        f'{indent}{var_name} = {_list_index_expr(src, params)}'
    )
    state.store(node.node_id, var_name)
    return True


def _emit_list_set_node(
    *,
    state: EmitState,
    node: AcherionNode,
    params: dict[str, Any],
    indent: str,
    var_name: str,
    method_owner_name: str | None = 'self',
) -> bool:
    del method_owner_name
    source_expr = _input_param_expr(
        params,
        'source',
        state.node_vars,
        fallback='[]',
    )
    value_expr = _input_param_expr(
        params,
        'value',
        state.node_vars,
    )
    base_var = f'{var_name}_base'
    state.lines.append(f'{indent}{base_var} = {source_expr}')
    state.lines.append(
        f'{indent}{var_name} = ('
        f'{base_var}.copy() if hasattr({base_var}, "copy") else '
        f'(list({base_var}) if isinstance({base_var}, (list, tuple, range)) '
        f'else [])'
        f')'
    )
    state.lines.append(f'{indent}try:')
    state.lines.append(
        f'{indent}    {_list_index_expr(var_name, params)} = {value_expr}'
    )
    state.lines.append(f'{indent}except (IndexError, TypeError, ValueError):')
    state.lines.append(f'{indent}    pass')
    state.store(node.node_id, var_name)
    return True


def _emit_dict_get_node(
    *,
    state: EmitState,
    node: AcherionNode,
    params: dict[str, Any],
    indent: str,
    var_name: str,
    method_owner_name: str | None = 'self',
) -> bool:
    del method_owner_name
    dict_expr = _input_param_expr(
        params,
        'source',
        state.node_vars,
    )
    key_expr = _input_param_expr(
        params,
        'key',
        state.node_vars,
        fallback="''",
    )
    default_expr = _input_param_expr(
        params,
        'default',
        state.node_vars,
        fallback='None',
    )
    dict_var = f'{var_name}_dict'
    state.lines.append(f'{indent}{dict_var} = {dict_expr}')
    state.lines.append(
        f'{indent}{var_name} = '
        f'{dict_var}.get({key_expr}, {default_expr}) '
        f'if isinstance({dict_var}, dict) else {default_expr}'
    )
    state.store(node.node_id, var_name)
    return True


def _emit_dict_set_node(
    *,
    state: EmitState,
    node: AcherionNode,
    params: dict[str, Any],
    indent: str,
    var_name: str,
    method_owner_name: str | None = 'self',
) -> bool:
    del method_owner_name
    dict_expr = _input_param_expr(
        params,
        'source',
        state.node_vars,
    )
    key_expr = _input_param_expr(
        params,
        'key',
        state.node_vars,
        fallback="''",
    )
    value_expr = _input_param_expr(
        params,
        'value',
        state.node_vars,
    )
    base_var = f'{var_name}_base'
    state.lines.append(f'{indent}{base_var} = {dict_expr}')
    state.lines.append(
        f'{indent}{var_name} = '
        f'dict({base_var}) if isinstance({base_var}, dict) else {{}}'
    )
    state.lines.append(f'{indent}{var_name}[{key_expr}] = {value_expr}')
    state.store(node.node_id, var_name)
    return True


def _emit_plot_figure_node(
    *,
    state: EmitState,
    node: AcherionNode,
    params: dict[str, Any],
    indent: str,
    var_name: str,
    method_owner_name: str | None = 'self',
) -> bool:
    del method_owner_name
    figure_type = str(params.get('figure_type') or 'scatter')
    named_sources = dict(params.get('named_sources') or {})
    figure_title = str(params.get('figure_title') or '').strip()
    trace_entry = _catalog_plotly.trace_entry(figure_type)
    if trace_entry is None:
        state.lines.append(
            f'{indent}{var_name} = None  # unknown figure type'
        )
        state.store(node.node_id, var_name)
        return True
    connected: dict[str, str] = {}
    for trace_param in trace_entry.params:
        src_id = str(named_sources.get(trace_param.name) or '').strip()
        if src_id:
            connected[trace_param.name] = state.source_expr(src_id)
    gauge_min = connected.pop('gauge_axis_range_min', None)
    gauge_max = connected.pop('gauge_axis_range_max', None)
    err_y = connected.pop('error_y_array', None)
    kw_parts: list[str] = []
    if figure_type == 'line' and 'mode' not in connected:
        kw_parts.append("mode='lines'")
    for param_name, expr in connected.items():
        kw_parts.append(f'{param_name}={expr}')
    if err_y is not None:
        kw_parts.append(f"error_y=dict(type='data', array={err_y})")
    if gauge_min is not None or gauge_max is not None:
        low = gauge_min or 'None'
        high = gauge_max or 'None'
        kw_parts.append(f'gauge=dict(axis=dict(range=[{low}, {high}]))')
    kw_str = ', '.join(kw_parts)
    state.lines.extend([
        f'{indent}{var_name} = go.Figure('
        f'data=[{trace_entry.go_class}({kw_str})])',
    ])
    if figure_title:
        state.lines.append(
            f"{indent}{var_name}.update_layout(title_text={figure_title!r})"
        )
    state.store(node.node_id, var_name)
    return True


_COMPUTE_NODE_EMITTERS: dict[str, _EmitNodeHandler] = {
    'constant': _emit_constant_node,
    'call_function': _emit_callable_node,
    'custom_function': _emit_callable_node,
    'call_method': _emit_call_method_node,
    'get_attribute': _emit_get_attribute_node,
    'set_attribute': _emit_set_attribute_node,
    'for_each': _emit_for_each_node,
    'collect': _emit_collect_node,
    'compare': _emit_compare_node,
    'branch_value': _emit_branch_value_node,
    'op_arithmetic': _emit_op_arithmetic_node,
    'op_unary': _emit_op_unary_node,
    'op_logic': _emit_op_logic_node,
    'op_not': _emit_op_not_node,
    'make_list': _emit_make_list_node,
    'make_dict': _emit_make_dict_node,
    'list_index': _emit_list_index_node,
    'list_set': _emit_list_set_node,
    'dict_get': _emit_dict_get_node,
    'dict_set': _emit_dict_set_node,
    'plot_figure': _emit_plot_figure_node,
}


def _emit_common_compute_node(
    *,
    state: EmitState,
    node: AcherionNode,
    params: dict[str, Any],
    indent: str,
    var_name: str,
    method_owner_name: str | None = 'self',
) -> bool:
    """Emit code for node kinds shared by helper and top-level compilers."""
    handler = _COMPUTE_NODE_EMITTERS.get(node.kind)
    if handler is None:
        return False
    return handler(
        state=state,
        node=node,
        params=params,
        indent=indent,
        var_name=var_name,
        method_owner_name=method_owner_name,
    )


"""Standalone event definitions for embedded Acherion runtimes."""

from __future__ import annotations

from collections.abc import Iterable, Mapping
from dataclasses import dataclass
import re
from typing import Any

EXTERNAL_EVENT_NODE_KIND = 'external_event'
RUN_EVENT_KEY = 'run'

_EVENT_IDENTIFIER_RE = re.compile(r'[^a-zA-Z0-9_]+')


@dataclass(frozen=True, slots=True)
class AcherionExternalEvent:
    """Describe one external event that a host can feed into Acherion."""

    event_key: str
    title: str
    handler_name: str
    description: str = ''
    source_kind: str = 'host'
    component_key: str = ''
    component_event: str = ''


def _clean_identifier(text: str, fallback: str = 'event') -> str:
    """Return one safe Python identifier fragment."""
    identifier = _EVENT_IDENTIFIER_RE.sub('_', str(text or '').strip().lower())
    identifier = re.sub(r'_+', '_', identifier).strip('_')
    if not identifier:
        identifier = fallback
    if identifier[0].isdigit():
        identifier = f'{fallback}_{identifier}'
    return identifier


def acherion_event_handler_name(
    event_key: str,
    *,
    prefix: str = 'on',
) -> str:
    """Return the default handler name for one external event key."""
    identifier = _clean_identifier(event_key)
    if not prefix:
        return identifier
    return f'{prefix}_{identifier}'


def component_event_key(component_key: str, component_event: str) -> str:
    """Return the stable runtime event key for one component event."""
    clean_component_key = _clean_identifier(component_key, 'component')
    clean_component_event = _clean_identifier(component_event, 'event')
    return (
        f'component:{clean_component_key}:{clean_component_event}'
    )


def build_component_external_event(
    *,
    component_key: str,
    component_event: str,
    component_label: str = '',
    event_label: str = '',
    description: str = '',
) -> AcherionExternalEvent:
    """Return one normalized component-driven external event."""
    clean_component_key = _clean_identifier(component_key, 'component')
    clean_component_event = _clean_identifier(component_event, 'event')
    clean_component_label = str(component_label or component_key).strip()
    clean_event_label = str(event_label or component_event).strip()
    event_key = component_event_key(
        clean_component_key,
        clean_component_event,
    )
    return AcherionExternalEvent(
        event_key=event_key,
        title=(
            f'Event: {clean_component_label or clean_component_key} - '
            f'{clean_event_label or clean_component_event}'
        ),
        handler_name=acherion_event_handler_name(event_key),
        description=str(description or '').strip(),
        source_kind='component',
        component_key=clean_component_key,
        component_event=clean_component_event,
    )


def acherion_external_event_params(
    event: AcherionExternalEvent,
) -> dict[str, str]:
    """Return one plain-dict payload for persisted/generated event metadata."""
    return {
        'event_key': event.event_key,
        'title': event.title,
        'handler_name': event.handler_name,
        'description': event.description,
        'source_kind': event.source_kind,
        'component_key': event.component_key,
        'component_event': event.component_event,
    }


def normalize_acherion_external_event_params(
    params: Mapping[str, Any] | None,
) -> dict[str, Any]:
    """Return cleaned params for one external-event node payload."""
    event_data = dict(params or {})
    event_key = str(event_data.get('event_key') or '').strip()
    handler_name = str(event_data.get('handler_name') or '').strip()
    description = str(event_data.get('description') or '').strip()
    component_key = str(event_data.get('component_key') or '').strip()
    component_event = str(event_data.get('component_event') or '').strip()
    source_kind = str(event_data.get('source_kind') or '').strip() or 'host'
    if component_key or component_event:
        source_kind = 'component'
    if not handler_name and event_key == RUN_EVENT_KEY:
        handler_name = 'run'
    elif not handler_name and event_key:
        handler_name = acherion_event_handler_name(event_key)
    event_data.update({
        'event_key': event_key,
        'handler_name': handler_name,
        'description': description,
        'source_kind': source_kind,
        'component_key': component_key,
        'component_event': component_event,
    })
    return event_data


def normalize_acherion_external_events(
    events: Mapping[str, Any] | None,
) -> dict[str, AcherionExternalEvent]:
    """Return a cleaned event mapping keyed by normalized event key."""
    normalized: dict[str, AcherionExternalEvent] = {}
    if not events:
        return normalized
    for raw_key, raw_event in events.items():
        if isinstance(raw_event, AcherionExternalEvent):
            event_data = acherion_external_event_params(raw_event)
        elif isinstance(raw_event, Mapping):
            event_data = normalize_acherion_external_event_params(raw_event)
        else:
            continue
        event_key = str(event_data.get('event_key') or raw_key).strip()
        title = str(event_data.get('title') or '').strip()
        handler_name = str(event_data.get('handler_name') or '').strip()
        description = str(event_data.get('description') or '').strip()
        component_key = str(event_data.get('component_key') or '').strip()
        component_event = str(event_data.get('component_event') or '').strip()
        source_kind = str(event_data.get('source_kind') or '').strip() or 'host'
        if not title:
            title = f'Event: {event_key}'
        if not event_key or not title or not handler_name:
            continue
        normalized[event_key] = AcherionExternalEvent(
            event_key=event_key,
            title=title,
            handler_name=handler_name,
            description=description,
            source_kind=source_kind,
            component_key=component_key,
            component_event=component_event,
        )
    return normalized


def resolve_acherion_external_event(
    events: Mapping[str, Any] | None,
    event_key: str,
) -> AcherionExternalEvent | None:
    """Return one normalized external event by key, if present."""
    return normalize_acherion_external_events(events).get(
        str(event_key or '').strip()
    )


def component_events_for_key(
    events: Mapping[str, Any] | None,
    component_key: str,
) -> tuple[AcherionExternalEvent, ...]:
    """Return all component-driven events bound to one component key."""
    clean_component_key = _clean_identifier(component_key, 'component')
    if not clean_component_key:
        return ()
    return tuple(
        event
        for event in normalize_acherion_external_events(events).values()
        if event.source_kind == 'component'
        and event.component_key == clean_component_key
    )


def collect_acherion_external_events(
    nodes: Iterable[Any],
) -> dict[str, AcherionExternalEvent]:
    """Return normalized external events declared by graph source nodes."""
    events: dict[str, dict[str, Any]] = {}
    for node in nodes:
        if str(getattr(node, 'kind', '') or '').strip() != EXTERNAL_EVENT_NODE_KIND:
            continue
        params = dict(getattr(node, 'params', {}) or {})
        if str(params.get('parent_function') or '').strip():
            continue
        event_data = normalize_acherion_external_event_params(params)
        event_key = str(event_data.get('event_key') or '').strip()
        if not event_key:
            continue
        title = str(getattr(node, 'title', '') or params.get('title') or '').strip()
        if title:
            event_data['title'] = title
        events[event_key] = event_data
    return normalize_acherion_external_events(events)


def serializable_acherion_external_events(
    events: Mapping[str, Any] | None,
) -> dict[str, dict[str, str]]:
    """Return one plain-dict mapping safe to inject into generated code."""
    return {
        event_key: acherion_external_event_params(event)
        for event_key, event in normalize_acherion_external_events(events).items()
    }


def default_acherion_external_events() -> dict[str, AcherionExternalEvent]:
    """Return builtin host-driven events available to generic graphs."""
    return normalize_acherion_external_events({
        RUN_EVENT_KEY: {
            'event_key': RUN_EVENT_KEY,
            'title': 'Event: Run',
            'handler_name': 'run',
            'description': 'Start run() execution for this graph path.',
        },
    })


def compose_acherion_external_events(
    *event_maps: Mapping[str, Any] | None,
    include_defaults: bool = True,
) -> dict[str, AcherionExternalEvent]:
    """Return normalized host events merged with optional builtin defaults."""
    merged: dict[str, Any] = {}
    if include_defaults:
        merged.update(default_acherion_external_events())
    for event_map in event_maps:
        if not event_map:
            continue
        merged.update(dict(event_map))
    return normalize_acherion_external_events(merged)


"""Function-box helper mixin for visual-logic graph ops."""

# pyright: reportAttributeAccessIssue=false

from __future__ import annotations

from typing import Any, Protocol, cast

from acherion.constants import _NODE_STEP
from acherion.model import AcherionNode, _default_node


class _FunctionBoxOwner(Protocol):
    def _node_by_id(self, node_id: str) -> AcherionNode | None:
        ...

    def _pure_node_id(self, source_id: str) -> str:
        ...

    def _function_parent_id(self, node: AcherionNode) -> str:
        ...


class _GraphOpsFunctionBoxesMixin:
    """Function-box traversal and port-sync helpers."""

    _selected_connection_id: str | None = None

    @staticmethod
    def _is_function_box(node: AcherionNode) -> bool:
        return node.kind == 'function_box'

    @staticmethod
    def _is_function_input(node: AcherionNode) -> bool:
        return node.kind == 'function_input'

    @staticmethod
    def _is_function_output(node: AcherionNode) -> bool:
        return node.kind == 'function_output'

    @staticmethod
    def _is_function_entry(node: AcherionNode) -> bool:
        return node.kind == 'function_entry'

    @staticmethod
    def _function_entry_node_id(box_id: str) -> str:
        return f'{box_id}:entry'

    def _function_parent_id(self: Any, node: AcherionNode) -> str:
        return str(node.params.get('parent_function') or '').strip()

    def _function_child_nodes(self: Any, box_id: str) -> list[AcherionNode]:
        return [
            node for node in self._manual_nodes()
            if self._function_parent_id(node) == box_id
        ]

    def _function_entry_node(self: Any, box_id: str) -> AcherionNode | None:
        for node in self._function_child_nodes(box_id):
            if self._is_function_entry(node):
                return cast(AcherionNode, node)
        return None

    def _visible_function_child_nodes(self: Any, box_id: str) -> list[AcherionNode]:
        return [
            node for node in self._function_child_nodes(box_id)
            if not self._is_function_input(node)
            and not self._is_function_output(node)
            and not self._is_function_entry(node)
        ]

    def _node_has_function_ancestor(
        self: Any,
        node_id: str,
        ancestor_box_id: str,
    ) -> bool:
        current = self._node_by_id(node_id)
        seen: set[str] = set()
        while current is not None:
            parent_id = self._function_parent_id(current)
            if not parent_id or parent_id in seen:
                return False
            if parent_id == ancestor_box_id:
                return True
            seen.add(parent_id)
            current = self._node_by_id(parent_id)
        return False

    def _containing_function_box_id(
        self: Any,
        x: int,
        y: int,
        *,
        exclude_node_id: str = '',
    ) -> str:
        x = self._world_to_canvas_x(x)
        y = self._world_to_canvas_y(y)
        candidates: list[tuple[int, int, int, str]] = []
        for node in self._manual_nodes():
            if not self._is_function_box(node):
                continue
            if exclude_node_id and node.node_id == exclude_node_id:
                continue
            if exclude_node_id and self._node_has_function_ancestor(
                node.node_id,
                exclude_node_id,
            ):
                continue
            left, top, width, height = self._node_bounds(node)
            if left <= x <= left + width and top <= y <= top + height:
                candidates.append((width * height, top, left, node.node_id))
        if not candidates:
            return ''
        candidates.sort()
        return candidates[0][3]

    def _assign_node_to_containing_function_box(
        self: Any,
        node: AcherionNode,
        left: int,
        top: int,
    ) -> None:
        if self._is_system_node(node):
            return
        if self._is_function_input(node) or self._is_function_output(node):
            return
        center_x = left + max(1, self._node_width(node) // 2)
        center_y = top + max(1, self._node_height(node) // 2)
        parent_box_id = self._containing_function_box_id(
            center_x,
            center_y,
            exclude_node_id=node.node_id,
        )
        node.params['parent_function'] = parent_box_id

    def _selected_function_box_target(self: Any) -> str:
        if not self._selected_node_ids:
            return ''
        target_ids: set[str] = set()
        for node in self._manual_nodes():
            if node.node_id not in self._selected_node_ids:
                continue
            if self._is_function_box(node):
                target_ids.add(node.node_id)
            else:
                target_ids.add(self._function_parent_id(node))
        target_ids.discard('')
        if len(target_ids) == 1:
            return next(iter(target_ids))
        return ''

    def _place_node_inside_function_box(
        self: Any,
        node: AcherionNode,
        box_id: str,
    ) -> None:
        box = self._node_by_id(box_id)
        if box is None or not self._is_function_box(box):
            return
        visible_children = [
            child for child in self._visible_function_child_nodes(box_id)
            if child.node_id != node.node_id
        ]
        if visible_children:
            node.params['x'] = max(
                self._node_world_left(box) + 220,
                max(
                    self._node_world_left(child)
                    for child in visible_children
                ),
            )
            node.params['y'] = max(
                self._node_world_top(box) + 116,
                max(
                    self._node_world_top(child)
                    for child in visible_children
                )
                + _NODE_STEP,
            )
        else:
            node.params['x'] = self._node_world_left(box) + 220
            node.params['y'] = self._node_world_top(box) + 116
        node.params['parent_function'] = box_id

    def _build_function_entry_node(
        self: Any,
        box: AcherionNode,
    ) -> AcherionNode:
        entry = _default_node('function_entry')
        entry.node_id = self._function_entry_node_id(box.node_id)
        entry.title = 'Entry'
        entry.params['parent_function'] = box.node_id
        entry.params['manual_position'] = True
        entry.params['dock'] = 'free'
        entry.params['x'] = self._node_world_left(box) + 48
        entry.params['y'] = self._node_world_top(box) + 116
        return entry

    def _ensure_function_box_entries(self: Any) -> None:
        existing_ids = {node.node_id for node in self._graph.nodes}
        new_entries: list[AcherionNode] = []
        for box in [node for node in self._manual_nodes() if self._is_function_box(node)]:
            entry = self._function_entry_node(box.node_id)
            expected_id = self._function_entry_node_id(box.node_id)
            if entry is not None:
                if entry.node_id != expected_id and expected_id not in existing_ids:
                    existing_ids.discard(entry.node_id)
                    entry.node_id = expected_id
                    existing_ids.add(expected_id)
                entry.title = 'Entry'
                entry.params['parent_function'] = box.node_id
                continue
            if expected_id in existing_ids:
                continue
            new_entry = self._build_function_entry_node(box)
            new_entries.append(new_entry)
            existing_ids.add(new_entry.node_id)
        if new_entries:
            self._graph.nodes.extend(new_entries)

    def _prune_obsolete_function_io_nodes(self: Any) -> None:
        removed_node_ids = {
            node.node_id
            for node in self._graph.nodes
            if self._is_function_input(node) or self._is_function_output(node)
        }
        if removed_node_ids:
            self._graph.nodes = [
                node
                for node in self._graph.nodes
                if node.node_id not in removed_node_ids
            ]
            self._clear_removed_source_refs(
                removed_node_ids=removed_node_ids,
                removed_source_ids=set(),
            )
        for node in self._graph.nodes:
            if self._is_function_box(node):
                self._prune_box_io_metadata(node)

    def _function_box_boundary_input_sources(self: Any, box_id: str) -> list[str]:
        sources: list[str] = []
        seen_sources: set[str] = set()
        for node in self._graph.nodes:
            if self._is_function_input(node) or self._is_function_output(node):
                continue
            if not self._node_has_function_ancestor(node.node_id, box_id):
                continue
            for pin in self._input_pin_specs(node):
                source_id = str(self._input_source_id(node, pin['pin_id']) or '')
                if not source_id or source_id in seen_sources:
                    continue
                source_node = self._node_by_id(self._pure_node_id(source_id))
                if source_node is None:
                    continue
                if source_node.node_id == box_id:
                    continue
                if self._node_has_function_ancestor(source_node.node_id, box_id):
                    continue
                seen_sources.add(source_id)
                sources.append(source_id)
        return sources

    def _function_box_boundary_output_sources(self: Any, box_id: str) -> list[str]:
        sources: list[str] = []
        seen_sources: set[str] = set()
        for node in self._graph.nodes:
            if self._is_function_input(node) or self._is_function_output(node):
                continue
            if self._node_has_function_ancestor(node.node_id, box_id):
                continue
            for pin in self._input_pin_specs(node):
                source_id = str(self._input_source_id(node, pin['pin_id']) or '')
                if not source_id or source_id in seen_sources:
                    continue
                source_node = self._node_by_id(self._pure_node_id(source_id))
                if source_node is None:
                    continue
                if not self._node_has_function_ancestor(
                    source_node.node_id,
                    box_id,
                ):
                    continue
                seen_sources.add(source_id)
                sources.append(source_id)
        return sources

    def _function_box_pin_type_from_source(self: Any, source_id: str) -> str:
        source_node = self._node_by_id(self._pure_node_id(source_id))
        if source_node is None:
            return 'any'
        pin_index = self._source_pin_index(source_id)
        specs = self._output_pin_specs(source_node)
        if pin_index < len(specs):
            return str(specs[pin_index].get('type') or 'any')
        return 'any'

    @staticmethod
    def _is_auto_function_port_label(label: str, prefix: str) -> bool:
        stripped = label.strip().lower()
        return not stripped or stripped == prefix or stripped.startswith(f'{prefix} ')

    def _sync_function_box_ports(self: Any) -> None:
        self._prune_obsolete_function_io_nodes()

    def _source_inside_function_box(
        self,
        source_id: str,
        box_id: str,
    ) -> bool:
        owner = cast(_FunctionBoxOwner, self)
        source_node = owner._node_by_id(owner._pure_node_id(source_id))
        if source_node is None:
            return False
        is_inside: bool = owner._function_parent_id(source_node) == box_id
        return is_inside

    def _handle_function_box_input_click(
        self: Any,
        box: AcherionNode,
        input_node_id: str,
    ) -> None:
        pending = str(self._pending_source_node_id or '')
        if not pending or pending == input_node_id:
            self._start_connection(input_node_id)
            return
        if self._source_inside_function_box(pending, box.node_id):
            self._update_hint(
                'Function-box input pins accept external sources only. '
                'Click the input pin first, then wire it into an inner node.'
            )
            return
        self._connect_input_pin(box, f'fin:{input_node_id}')

    def _handle_function_box_output_click(
        self: Any,
        box: AcherionNode,
        output_node_id: str,
    ) -> None:
        output_node = self._node_by_id(output_node_id)
        if output_node is None:
            return
        pending = str(self._pending_source_node_id or '')
        external_source_id = self._box_output_source_id(box, output_node_id)
        if not pending or pending == external_source_id:
            self._start_connection(external_source_id)
            return
        if not self._source_inside_function_box(pending, box.node_id):
            self._update_hint(
                'Function-box output pins capture internal sources only. '
                'Select a node inside the box, or click the output pin '
                'first for external wiring.'
            )
            return
        self._connect_input_pin(output_node, 'source')

    def _box_output_source_id(self: Any, box: AcherionNode, output_node_id: str) -> str:
        output_order = list(box.params.get('output_order') or [])
        try:
            pin_index = output_order.index(output_node_id)
        except ValueError:
            pin_index = 0
        return f'{box.node_id}@{pin_index}'

    def _prune_box_io_metadata(self: Any, box: AcherionNode) -> None:
        box.params.pop('input_order', None)
        box.params.pop('output_order', None)
        box.params.pop('box_input_sources', None)


"""Shared constants for visual-logic render mixins."""

from __future__ import annotations

_DEFAULT_NODE_HEIGHT = 108
_NODE_BODY_GAP = 8
_FUNCTION_BOX_TOP_PAD = 126
_FUNCTION_BOX_SIDE_PAD = 176
_FUNCTION_BOX_SIDE_PAD_COMPACT = 40
_FUNCTION_BOX_BOTTOM_PAD = 40
_FUNCTION_BOX_MIN_WIDTH = 720
_FUNCTION_BOX_MIN_WIDTH_COMPACT = 440
_FUNCTION_BOX_MIN_HEIGHT = 260
_FUNCTION_BOX_PORT_TOP = 122
_FUNCTION_BOX_PORT_CARD_HEIGHT = 40
_FUNCTION_BOX_PORT_GAP = 8
_FUNCTION_BOX_PORT_BOTTOM_PAD = 24
_GROUP_FRAME_SIDE_PAD = 28
_GROUP_FRAME_TOP_PAD = 46
_GROUP_FRAME_BOTTOM_PAD = 28

FUNCTION_PORT_TYPE_OPTIONS: dict[str, str] = {
    'any': 'any',
    'float': 'float',
    'bool': 'bool',
    'str': 'str',
    'list': 'list',
    'np.ndarray': 'np.ndarray',
    'object': 'object',
    'Figure': 'Figure',
}